#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from .TextEntry import TextEntry
import os
import subprocess
from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib  # type: ignore


class StartupWmClassEntry(TextEntry):
    def __init__(self, property_name, use_headerbar):
        super().__init__(property_name=property_name)

        xprop = GLib.find_program_in_path("xprop")
        if xprop is not None:
            self.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, 'edit-find-symbolic')
            self.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _('Identify Window')
            )
            self.connect(
                'icon-press', self._on_icon_press, xprop, use_headerbar)

    def _on_icon_press(self, entry, icon, event, xprop, use_headerbar):
        dialog = XpropWindowDialog(self.get_toplevel(), xprop, use_headerbar)
        wm_classes = dialog.run_xprop()
        current = entry.get_text()
        for wm_class in wm_classes:
            if wm_class != current:
                self.set_value(wm_class)
                return


class XpropWindowDialog(Gtk.MessageDialog):
    def __init__(self, parent, xprop_binary, use_headerbar):
        # Translators: Identify Window Dialog, primary text.
        primary = _("Identify Window")
        # Translators: Identify Window Dialog, secondary text. The selected
        # application is displayed in the placeholder text.
        secondary = _(
            "Click on the main application window for this launcher.")

        Gtk.MessageDialog.__init__(self, transient_for=parent, modal=True,
                                   message_type=Gtk.MessageType.INFO,
                                   buttons=Gtk.ButtonsType.OK,
                                   text=primary, use_header_bar=False)
        self.format_secondary_markup(secondary)

        self.binary = xprop_binary
        self.process = None
        self.classes = []

    def run_xprop(self):
        GLib.timeout_add(500, self.start_xprop)
        self.run()
        self.classes.sort()
        return self.classes

    def start_xprop(self):
        cmd = [self.binary, 'WM_CLASS']
        self.classes = []
        env = os.environ.copy()
        self.process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        GLib.idle_add(self.check_xprop)
        return False

    def check_xprop(self):
        if self.process is None:
            return True
        if self.process.poll() is None:
            return True
        if self.process.stdout is None:
            return True
        output = self.process.stdout.read().decode('UTF-8').strip()
        if output.startswith("WM_CLASS"):
            values = output.split("=", 1)[1].split(", ")
            for value in values:
                value = value.strip()
                value = value[1:-1]
                if value not in self.classes:
                    self.classes.append(value)
        self.destroy()
        return False



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import locale
import os
import json
from locale import gettext as _

import subprocess

from gi.repository import GLib  # type: ignore

locale.textdomain('menulibre')


sudo = os.getuid() == 0
default_locale = locale.getdefaultlocale()[0]


class MenulibreDesktopEntry:

    """Basic class for Desktop Entry files"""

    def __init__(self, filename=None):
        """Initialize the MenulibreDesktopEntry instance."""
        self.keyfile = GLib.KeyFile.new()
        if filename is not None and os.path.isfile(filename):
            self.load_properties(filename)
        else:
            self['Version'] = '1.1'
            self['Type'] = 'Application'
            self['Name'] = _('New Launcher')
            self['Comment'] = ''
            self['Icon'] = 'applications-other'
            self['Exec'] = ''
            self['Path'] = ''
            self['Terminal'] = 'false'
            self['StartupNotify'] = 'false'
            self['Categories'] = ''

    def __getitem__(self, prop_name):
        """Get property from this object like a dictionary."""
        return self.get_property('Desktop Entry', prop_name, default_locale)

    def __setitem__(self, key, value):
        """Set property to this object like a dictionary."""
        self._set_value("Desktop Entry", key, value)
        if key in ["Name", "GenericName", "Comment", "Keywords"]:
            self._set_locale_string("Desktop Entry", key, default_locale,
                                    value)

    def load_properties(self, filename):
        """Load the properties."""
        self.keyfile = GLib.KeyFile.new()
        self.keyfile.load_from_file(filename,
                                    GLib.KeyFileFlags.KEEP_TRANSLATIONS)

    def get_property(self, category, prop_name, locale_str=default_locale):
        """Return the value of the specified property."""
        prop = self.get_named_property(category, prop_name, locale_str)
        if prop in ['true', 'false']:
            return prop == 'true'
        if prop_name in ['Hidden', 'NoDisplay', 'Terminal', 'StartupNotify']:
            return False
        return prop

    def get_named_property(self, group, key, locale_str=None):
        """Return the value of the specified named property."""
        if key in ["Name", "GenericName", "Comment", "Keywords"]:
            if locale_str is not None:
                return self._get_locale_string(group, key, locale_str)

        value = self._get_value(group, key)

        if value is not None:
            return value

        return ""

    def get_actions(self):
        """Return a list of the Unity action groups."""
        if "Actions" in self._get_keys("Desktop Entry"):
            action_key = "Actions"
        elif "X-Ayatana-Desktop-Shortcuts" in self._get_keys("Desktop Entry"):
            action_key = "X-Ayatana-Desktop-Shortcuts"
        else:
            return json.dumps([])

        enabled_quicklists = self._get_string_list("Desktop Entry", action_key)

        quicklists = []

        for group in self._get_groups():
            name = self._get_action_group_name(group)
            if name is None:
                continue
            displayed_name = self.get_property(group, "Name")
            command = self.get_property(group, "Exec")
            enabled = name in enabled_quicklists
            quicklists.append([enabled, name, displayed_name, command])

        return json.dumps(quicklists)

    def _get_action_group_name(self, group):
        if group.startswith("Desktop Action "):
            name = group.replace("Desktop Action", "")
        elif group.endswith(" Shortcut Group"):
            name = group.replace("Shortcut Group", "")
        else:
            return None
        name = name.strip()
        if len(name) > 0:
            return name
        return None

    def _get_locale_string(self, group, key, locale_str):
        try:
            value = self.keyfile.get_locale_string(group, key, locale_str)
        except GLib.Error:
            value = None

        if value is not None:
            return value

        if '_' in locale_str:
            locale_str = locale_str.split("_")[0]
            return self._get_locale_string(group, key, locale_str)

        return self._get_string(group, key)

    def _set_locale_string(self, group, key, locale_str, value):
        self.keyfile.set_locale_string(group, key, locale_str, value)

    def _get_string(self, group, key):
        try:
            value = self.keyfile.get_string(group, key)
        except GLib.Error:
            value = None
        if value is not None:
            return value
        return ""

    def _get_value(self, group, key):
        try:
            value = self.keyfile.get_value(group, key)
        except GLib.Error:
            value = None
        return value

    def _set_value(self, group, key, value):
        self.keyfile.set_value(group, key, value)

    def _get_string_list(self, group, key):
        try:
            value = self.keyfile.get_string_list(group, key)
        except GLib.Error:
            value = None
        if value is not None:
            return value
        return []

    def _get_groups(self):
        try:
            return self.keyfile.get_groups()[0]
        except GLib.Error:
            return []

    def _get_keys(self, group):
        try:
            return self.keyfile.get_keys(group)[0]
        except GLib.Error:
            return []


def desktop_menu_update():
    subprocess.call(["xdg-desktop-menu", "forceupdate"])


def desktop_menu_install(directory_files, desktop_files):
    """Install one or more applications in a submenu of the desktop menu
    system.  If multiple directory files are provided each file will represent
    a submenu within the menu that precedes it, creating a nested menu
    hierarchy (sub-sub-menus). The menu entries themselves will be added to
    the last submenu. """
    # Check for the minimum required arguments
    if len(directory_files) == 0 or len(desktop_files) == 0:
        return True

    # Do not install to system paths.
    for path in GLib.get_system_config_dirs():
        for filename in directory_files:
            if filename.startswith(path):
                return True

    # xdg-desktop-menu doesn't behave nicely with vendor- directories
    # without vendor- applications.
    directory_dir = os.path.join(
        GLib.get_user_data_dir(),
        "desktop-directories")
    for filename in directory_files:
        if not filename.startswith(directory_dir):
            continue
        if "/" not in filename:
            continue
        basename = os.path.basename(filename)
        relative = filename.split("desktop-directories/")[1]
        vendor_path = relative.replace("/%s" % basename, "")
        if len(vendor_path) == 0:
            continue
        vendor_prefix = vendor_path.replace("/", "-") + "-"
        if basename.startswith(vendor_prefix):
            return False

    cmd_list = ["xdg-desktop-menu", "install", "--novendor"]
    cmd_list = cmd_list + directory_files + desktop_files
    subprocess.call(cmd_list)

    return True


def desktop_menu_uninstall(directory_files, desktop_files):  # noqa
    """Remove applications or submenus from the desktop menu system
    previously installed with xdg-desktop-menu install."""
    # Check for the minimum required arguments
    if len(directory_files) == 0 or len(desktop_files) == 0:
        return

    # Do not uninstall from system paths.
    for path in GLib.get_system_config_dirs():
        for filename in directory_files:
            if filename.startswith(path):
                return

    # xdg-desktop-menu uninstall does not work... implement ourselves.
    basenames = []
    for filename in directory_files:
        basenames.append(os.path.basename(filename))
    basenames.sort()
    base_filename = os.path.basename(desktop_files[0])

    # Find the file with all the details to remove the filename.
    merged_dir = os.path.join(GLib.get_user_config_dir(),
                              "menus", "applications-merged")

    for filename in os.listdir(merged_dir):
        filename = os.path.join(merged_dir, filename)
        found_directories = []
        filename_found = False
        with open(filename, 'r') as open_file:
            write_contents = ""
            for line in open_file:
                if "<Filename>" in line:
                    if base_filename in line:
                        filename_found = True
                    else:
                        write_contents += line
                else:
                    write_contents += line
                if "<Directory>" in line:
                    line = line.split("<Directory>")[1]
                    line = line.split("</Directory>")[0]
                    found_directories.append(line)
        if filename_found:
            found_directories.sort()
            if basenames == found_directories:
                with open(filename, 'w') as open_file:
                    open_file.write(write_contents)
                return



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = [
    'project_path_not_found',
    'get_data_file',
    'get_data_path',
]

# Where your project will look for your data (for instance, images and ui
# files). By default, this is ../data, relative your trunk layout
__menulibre_data_directory__ = '/usr/local/share/menulibre/'
__license__ = 'GPL-3'
__version__ = '2.4.0'

import os


class project_path_not_found(Exception):
    """Raised when we can't find the project directory."""


def get_data_file(*path_segments):
    """Get the full path to a data file.

    Returns the path to a file underneath the data directory (as defined by
    `get_data_path`). Equivalent to os.path.join(get_data_path(),
    *path_segments).
    """
    return os.path.join(get_data_path(), *path_segments)


def get_data_path():
    """Retrieve menulibre data path

    This path is by default <menulibre_lib_path>/../data/ in trunk
    and /usr/share/menulibre in an installed version but this path
    is specified at installation time.
    """
    if __menulibre_data_directory__ == '../data/':
        path = os.path.join(
            os.path.dirname(__file__), __menulibre_data_directory__)
        abs_data_path = os.path.abspath(path)
    else:
        abs_data_path = os.path.abspath(__menulibre_data_directory__)
    if not os.path.exists(abs_data_path):
        print(abs_data_path)
        raise project_path_not_found

    return abs_data_path


def get_version():
    """Retrieve the program version."""
    return __version__



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # type: ignore


class SwitchEntry(Gtk.Switch):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, property_name):
        super().__init__()
        self._property_name = property_name

        self.set_halign(Gtk.Align.END)
        self.set_valign(Gtk.Align.CENTER)
        self.set_margin_end(1)

        self.connect('notify::active', self._on_changed)

    def set_value(self, value):
        self.set_active(value)

    def get_value(self):
        return self.get_active()

    def _on_changed(self, status, widget):
        self.emit('value-changed', self._property_name, self.get_value())



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # type: ignore


class Toolbar(Gtk.Toolbar):

    def __init__(self):
        super().__init__()
        self.set_show_arrow(False)

        context = self.get_style_context()
        context.add_class("primary-toolbar")

    def add_menu_button(self, icon_name, label, menu):
        item = Gtk.ToolItem.new()

        image = Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.LARGE_TOOLBAR)
        image.set_pixel_size(24)

        button = Gtk.MenuButton.new()
        button.set_menu_model(menu)
        button.set_use_popover(False)
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.set_tooltip_text(label)
        button.add(image)

        context = button.get_style_context()
        context.add_class("toolbutton")

        item.add(button)
        self.add(item)

        return item

    def add_separator(self):
        separator = Gtk.SeparatorToolItem.new()
        self.add(separator)
        return separator

    def add_button(self, icon_name, label):
        image = Gtk.Image.new_from_icon_name(
            icon_name, Gtk.IconSize.LARGE_TOOLBAR)
        image.set_pixel_size(24)

        item = Gtk.ToolButton.new(image, label)
        item.set_tooltip_text(label)

        self.add(item)

        return item

    def add_search(self, widget):
        item = Gtk.ToolItem.new()

        item.add(widget)
        self.add(item)

        return item



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject  # type: ignore

import logging
logger = logging.getLogger('menulibre')


class History(GObject.GObject):
    """The MenulibreHistory object. This stores all history for Menulibre and
    allows for Undo/Redo/Revert functionality."""

    __gsignals__ = {
        'undo-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
                         (GObject.TYPE_BOOLEAN,)),
        'redo-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
                         (GObject.TYPE_BOOLEAN,)),
        'revert-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
                           (GObject.TYPE_BOOLEAN,))
    }

    def __init__(self):
        """Initialize the MenulibreHistory object."""
        GObject.GObject.__init__(self)
        self._undo = []
        self._redo = []
        self._restore = dict()
        self._block = False

    def append(self, key, before, after):
        """Add a new change to the History, clear the redo."""
        if self._block:
            return
        self._append_undo(key, before, after)
        self._clear_redo()
        self._check_revert()

    def store(self, key, value):
        """Store an original value to be used for reverting."""
        self._restore[key] = value

    def restore(self):
        """Return a copy of the restore dictionary."""
        return self._restore.copy()

    def undo(self):
        """Return the next key-value pair to undo, push it to redo."""
        key, before, after = self._pop_undo()
        self._append_redo(key, before, after)
        self._check_revert()
        return (key, before)

    def redo(self):
        """Return the next key-value pair to redo, push it to undo."""
        key, before, after = self._pop_redo()
        self._append_undo(key, before, after)
        self._check_revert()
        return (key, after)

    def clear(self):
        """Clear all history items."""
        self._clear_undo()
        self._clear_redo()
        self._restore.clear()
        self._check_revert()

    def block(self):
        """Block all future history changes."""
        logger.debug('Blocking history updates')
        self._block = True

    def unblock(self):
        """Unblock all future history changes."""
        logger.debug('Unblocking history updates')
        self._block = False

    def is_blocked(self):
        """Is History allowed currently?"""
        return self._block

    def _append_undo(self, key, before, after):
        """Internal append_undo function. Emit 'undo-changed' if the undo stack
        now contains a history."""
        self._undo.append((key, before, after))
        if len(self._undo) == 1:
            self.emit('undo-changed', True)

    def _pop_undo(self):
        """Internal pop_undo function. Emit 'undo-changed' if the undo stack is
        now empty."""
        history = self._undo.pop()
        if len(self._undo) == 0:
            self.emit('undo-changed', False)
        return history

    def _clear_undo(self):
        """Internal clear_undo function. Emit 'undo-changed' if the undo stack
        previously had items."""
        has_history = len(self._undo) > 0
        self._undo.clear()
        if has_history:
            self.emit('undo-changed', False)

    def _clear_redo(self):
        """Internal clear_redo function. Emit 'redo-changed' if the redo stack
        previously had items."""
        has_history = len(self._redo) > 0
        self._redo.clear()
        if has_history:
            self.emit('redo-changed', False)

    def _append_redo(self, key, before, after):
        """Internal append_redo function. Emit 'redo-changed' if the redo stack
        now contains a history."""
        self._redo.append((key, before, after))
        if len(self._redo) == 1:
            self.emit('redo-changed', True)

    def _pop_redo(self):
        """Internal pop_redo function. Emit 'redo-changed' if the redo stack is
        now empty."""
        history = self._redo.pop()
        if len(self._redo) == 0:
            self.emit('redo-changed', False)
        return history

    def _check_revert(self):
        """Check if revert should now be enabled and emit the 'revert-changed'
        signal."""
        if len(self._undo) == 0 and len(self._redo) == 0:
            self.emit('revert-changed', False)
        elif len(self._undo) == 1 or len(self._redo) == 1:
            self.emit('revert-changed', True)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import shlex
from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio, GObject, GLib, Pango, Gdk  # type: ignore


class CommandEditorDialog(Gtk.Dialog):

    def __init__(self, parent, initial_text, use_header_bar):
        super().__init__(title=_("Command Editor"), transient_for=parent,
                         use_header_bar=use_header_bar, flags=0)  # type: ignore
        self.add_buttons(
            _('Cancel'), Gtk.ResponseType.CANCEL,
            _('Apply'), Gtk.ResponseType.OK
        )

        self.set_default_size(400, 200)
        self.set_default_response(Gtk.ResponseType.OK)

        box = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=18)
        box.set_margin_top(9)
        box.set_margin_bottom(12)
        box.set_margin_start(12)
        box.set_margin_end(12)

        command_box = Gtk.Box.new(
            orientation=Gtk.Orientation.VERTICAL, spacing=6)
        box.pack_start(command_box, False, False, 0)

        entry = CommandEditorEntry(initial_text)
        command_box.pack_start(entry, False, False, 0)

        toolbox = CommandToolBox()
        toolbox.connect('env-var-activated', self.on_env_var_activated, entry)
        toolbox.connect('command-activated', self.on_command_activated, entry)
        toolbox.connect('code-activated', self.on_code_activated, entry)
        command_box.pack_start(toolbox, False, False, 0)

        grid = Gtk.Grid()
        grid.set_row_spacing(3)
        grid.set_column_spacing(6)
        box.pack_start(grid, False, False, 0)

        self.state = {
            'env': entry.get_state('env'),
            'cmd': entry.get_state('cmd'),
            'field': entry.get_state('field')
        }

        env_hint = CommandEditorHint(grid, 0)
        env_hint.set_state(self.state['env'])

        cmd_hint = CommandEditorHint(grid, 1)
        cmd_hint.set_state(self.state['cmd'])

        fields_hint = CommandEditorHint(grid, 2)
        fields_hint.set_state(self.state['field'])

        entry.connect('env-state-changed',
                      self.on_state_changed, 'env', env_hint)
        entry.connect('cmd-state-changed',
                      self.on_state_changed, 'cmd', cmd_hint)
        entry.connect('field-state-changed',
                      self.on_state_changed, 'field', fields_hint)

        help = HelpButton(
            "https://github.com/bluesabre/menulibre/wiki/Recognized-Desktop-Entry-Keys#exec",
            _("Help"))
        box.pack_start(help, False, False, 0)

        self.get_content_area().add(box)
        self.show_all()

        self.text = ""

        self.connect("response", self.on_response, entry)

    def on_env_var_activated(self, widget, variable, value, entry):
        entry.insert_env_var(variable, value)

    def on_command_activated(self, widget, value, entry):
        entry.insert_command(value)

    def on_code_activated(self, widget, value, entry):
        entry.insert_field_code(value)

    def on_state_changed(self, widget, success, message, key, hint):
        self.state[key] = (success, message)
        hint.set_state(self.state[key])

        for key in list(self.state.keys()):
            if not self.state[key][0]:
                self.set_response_sensitive(Gtk.ResponseType.OK, False)
                return

        self.set_response_sensitive(Gtk.ResponseType.OK, True)

    def on_response(self, dialog, response_id, entry):
        if response_id == Gtk.ResponseType.OK:
            self.text = entry.get_text().strip()

    def get_text(self):
        return self.text


class HelpButton(Gtk.LinkButton):
    def __init__(self, uri, label):
        super().__init__(uri=uri)

        self.set_label(label)
        self.set_name('helpbutton')
        self.set_halign(Gtk.Align.START)


class CommandEditorEntry(Gtk.Box):
    __gsignals__ = {
        'env-state-changed': (GObject.SignalFlags.RUN_FIRST, None, (bool, str)),
        'cmd-state-changed': (GObject.SignalFlags.RUN_FIRST, None, (bool, str)),
        'field-state-changed': (GObject.SignalFlags.RUN_FIRST, None, (bool, str)),
    }

    def __init__(self, initial_text):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        context = self.get_style_context()
        context.add_class("linked")

        self.entry = Gtk.Entry()
        self.entry.set_text("delete.me")
        self.pack_start(self.entry, True, True, 0)

        revert_button = Gtk.Button.new_from_icon_name(
            "document-revert", Gtk.IconSize.BUTTON)
        revert_button.set_tooltip_text(_("Revert"))
        revert_button.connect("clicked", self.on_revert_button_clicked)
        self.pack_start(revert_button, False, False, 0)

        clear_button = Gtk.Button.new_from_icon_name(
            "edit-clear", Gtk.IconSize.BUTTON)
        clear_button.set_tooltip_text(_("Clear"))
        clear_button.connect("clicked", self.on_clear_button_clicked)
        self.pack_start(clear_button, False, False, 0)

        self.entry.connect("changed", self.on_entry_changed,
                           revert_button, clear_button)

        self.state = {
            'env': (True, ""),
            'cmd': (True, ""),
            'field': (True, ""),
        }

        self.text = initial_text
        self.entry.set_text(initial_text)

        self.show_all()

    def get_text(self):
        return self.entry.get_text()

    def set_text(self, text):
        self.entry.set_text(text)

    def get_state(self, key):
        return self.state[key]

    def insert_text(self, text):
        pos = self.entry.get_position()
        self.entry.insert_text(text, pos)
        self.entry.set_position(pos + len(text))

    def store_text(self, text):
        self.text = text

    def revert(self):
        self.set_text(self.text)

    def clear(self):
        self.set_text("")

    def on_revert_button_clicked(self, button):
        self.revert()
        self.entry.grab_focus()
        self.entry.set_position(len(self.text))

    def on_clear_button_clicked(self, button):
        self.clear()
        self.entry.grab_focus()

    def insert_env_var(self, var, val):
        if " " in val:
            val = shlex.quote(val)

        varval = "%s=%s" % (var, val)

        current = shlex.split(self.entry.get_text())

        if "env" not in current:
            new = ["env", varval] + current

        else:
            added = False
            new = []
            for item in current:
                if item == "env" or "=" in item or added:
                    new.append(item)
                elif not added:
                    new.append(varval)
                    new.append(item)
                    added = True
            if not added:
                new.append(varval)

        command = shlex.join(new)
        position = command.find(varval) + len(varval)

        self.entry.set_text(shlex.join(new))
        self.entry.set_position(position)

    def insert_component(self, value):
        if len(value) == 0:
            return

        text = self.entry.get_text()
        pos = self.entry.get_position()

        is_equals = pos > 0 and text[pos - 1] == "="
        need_space_before = pos > 0 and text[pos - 1] != "="
        need_space_after = len(text) > pos

        if is_equals and " " in value:
            value = shlex.quote(value)

        if need_space_before:
            value = " " + value

        if need_space_after:
            value = value + " "

        self.insert_text(value)

    def insert_command(self, value):
        self.insert_component(value)

    def insert_field_code(self, value):
        self.insert_component(value)

    def validate_field_codes(self, text):
        deprecated = ["%d", "%D", "%n", "%N", "%v", "%m"]
        singleton = ["%f", "%u", "%F", "%U"]
        desktop = ["%i", "%c", "%k"]
        keys = deprecated + singleton + desktop

        found = {}
        quoted = {}

        for key in keys:
            found[key] = 0
            quoted[key] = 0

            q = re.compile(key)
            found[key] = len(re.findall(q, text))

            q = re.compile('\"%s\"' % key)
            quoted[key] = len(re.findall(q, text))

            q = re.compile('\'%s\'' % key)
            quoted[key] += len(re.findall(q, text))

        single = 0
        for key in singleton:
            single += found[key]

        if single > 1:
            return _(
                "A single command line may only contain one of %f, %u, %F, or %U")

        for key in keys:
            if quoted[key] > 0:
                return _(
                    "Field code '%s' can not be used inside a quoted argument") % key

        for key in deprecated:
            if found[key] > 0:
                return _("Field code '%s' has been deprecated") % key

        return None

    def is_env_var(self, text):
        if "=" not in text:
            return False
        k = text.split("=")[0]
        return k == k.upper()

    def validate_cmd(self, text):
        results = {
            "env": [],
            "cmd": None,
            "args": [],
            "errors": {
                "env": None,
                "cmd": _("No command found"),
                "fields": None
            }
        }

        try:
            parts = shlex.split(text)
        except ValueError:
            return self._last_results

        if len(parts) == 0:
            return results

        env_found = parts[0] == "env"
        non_env_found = False

        for part in parts:
            if part == "env":
                continue

            if self.is_env_var(part):
                if not env_found:
                    results["errors"]["env"] = \
                        _("Environment variables should be preceded by env")
                elif non_env_found:
                    results["errors"]["env"] = \
                        _("Environment variables should precede the command")
                else:
                    results["env"].append(part)
                    continue
            else:
                non_env_found = True

            if results["cmd"] is None:
                results["cmd"] = part
            else:
                results["args"].append(part)

        if results["cmd"] is None:
            results["errors"]["cmd"] = _("No command found")
        elif results["cmd"].startswith("/") and not os.path.isfile(results["cmd"]):
            results["errors"]["cmd"] = \
                _("File '%s' not found") % results["cmd"]
        elif GLib.find_program_in_path(results["cmd"]) is None:
            results["errors"]["cmd"] = \
                _("Command '%s' not found") % results["cmd"]
        else:
            results["errors"]["cmd"] = None

        results["errors"]["fields"] = self.validate_field_codes(text)

        self._last_results = results

        return results

    def on_entry_changed(self, widget, revert_button, clear_button):
        text = widget.get_text()

        revert_button.set_sensitive(text != self.text)
        clear_button.set_sensitive(len(text) > 0)

        context = self.validate_cmd(text)

        if context["errors"]["env"] is not None:
            env = (False, context["errors"]["env"])
        else:
            env = (True, _('No environment variable errors'))

        if context["errors"]["cmd"] is not None:
            cmd = (False, context["errors"]["cmd"])
        else:
            cmd = (True, _('No command errors'))

        if context["errors"]["fields"] is not None:
            field = (False, context["errors"]["fields"])
        else:
            field = (True, _('No invalid field codes'))

        if env != self.state['env']:
            self.state['env'] = env  # type: ignore
            self.emit('env-state-changed', env[0], env[1])

        if cmd != self.state['cmd']:
            self.state['cmd'] = cmd  # type: ignore
            self.emit('cmd-state-changed', cmd[0], cmd[1])

        if field != self.state['field']:
            self.state['field'] = field  # type: ignore
            self.emit('field-state-changed', field[0], field[1])


class CommandToolBox(Gtk.Box):
    __gsignals__ = {
        'env-var-activated': (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        'command-activated': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        'code-activated': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        label = Gtk.Label.new("")
        label.set_markup("<b>%s</b>" % _("Add"))
        self.pack_start(label, False, False, 0)

        box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        context = box.get_style_context()
        context.add_class("linked")

        popover = CommandEditorEnvEditor()
        popover.connect("activated", self.on_env_var_activated)
        box.pack_start(popover, False, False, 0)

        popover = CommandEditorCommandEntry()
        popover.connect("activated", self.on_command_activated)
        box.pack_start(popover, False, False, 0)

        size_group = Gtk.SizeGroup.new(Gtk.SizeGroupMode.BOTH)

        popover = FieldCodeMenu(_("File"))
        popover.add_option("%f", _("A single file or directory."), size_group)
        popover.add_option(
            "%F", _("A list of files or directories."), size_group)
        popover.connect("row-activated", self.on_field_code_activated)
        box.pack_start(popover, False, False, 0)

        popover = FieldCodeMenu(_("URL"))
        popover.add_option("%u", _("A single URL."), size_group)
        popover.add_option("%U", _("A list of URLs."), size_group)
        popover.connect("row-activated", self.on_field_code_activated)
        box.pack_start(popover, False, False, 0)

        popover = FieldCodeMenu(_("Extra"))
        popover.add_option(
            "%i", _("The <tt>Icon</tt> key of this launcher."), size_group)
        popover.add_option(
            "%c",
            _("The translated <tt>Name</tt> key of this launcher."),
            size_group)
        popover.add_option(
            "%k", _("The file path of this launcher."), size_group)
        popover.connect("row-activated", self.on_field_code_activated)
        box.pack_start(popover, False, False, 0)

        self.add(box)

        self.show_all()

    def on_env_var_activated(self, widget, variable, value):
        self.emit('env-var-activated', variable, value)

    def on_command_activated(self, widget, value):
        self.emit('command-activated', value)

    def on_field_code_activated(self, widget, value):
        self.emit('code-activated', value)


class CommandEditorEnvEditor(Gtk.MenuButton):
    __gsignals__ = {
        'activated': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self):
        super().__init__()

        self.set_label(_("Environment variable"))
        self.set_mode(True)

        popover = Gtk.Popover(relative_to=self)
        popover.set_constrain_to(Gtk.PopoverConstraint.NONE)

        grid = Gtk.Grid()
        grid.set_row_spacing(6)
        grid.set_column_spacing(6)
        grid.set_border_width(6)

        label = Gtk.Label.new("")
        label.set_markup("<b>%s</b>" % _("Variable"))
        label.set_xalign(0.00)
        var_entry = Gtk.Entry()

        grid.attach(label, 0, 0, 1, 1)
        grid.attach(var_entry, 1, 0, 1, 1)
        var_entry.grab_focus()

        label = Gtk.Label.new("")
        label.set_markup("<b>%s</b>" % _("Value"))
        label.set_xalign(0.00)
        val_entry = Gtk.Entry()

        var_entry.connect("changed", self.on_var_entry_changed, val_entry)
        val_entry.connect("changed", self.on_val_entry_changed)

        var_entry.connect('activate', self.on_activate,
                          var_entry, val_entry, popover)
        val_entry.connect('activate', self.on_activate,
                          var_entry, val_entry, popover)

        popover.connect('closed', self.on_closed, var_entry, val_entry)

        grid.attach(label, 0, 1, 1, 1)
        grid.attach(val_entry, 1, 1, 1, 1)

        grid.show_all()

        popover.add(grid)

        self.set_popover(popover)
        self.show_all()

    def on_var_entry_changed(self, var_entry, val_entry):
        text = var_entry.get_text()
        text = text.lstrip()

        if "=" in text:
            parts = text.split("=", 2)
            if len(parts[1]) > 0:
                val_entry.set_text(parts[1].rstrip())
            text = parts[0].strip()

        text = text.upper()
        var_entry.set_text(text)

        if len(text) == 0:
            var_entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, None)
            var_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, None)
        elif " " in text:
            var_entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, "dialog-error")
            var_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY,
                _("Spaces not permitted in environment variables"))
        else:
            var_entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, "gtk-apply")
            var_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, None)

    def on_val_entry_changed(self, val_entry):
        text = val_entry.get_text()
        text = text.strip()

        if len(text) == 0:
            val_entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, None)
            val_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, None)
        else:
            val_entry.set_icon_from_icon_name(
                Gtk.EntryIconPosition.SECONDARY, "gtk-apply")
            val_entry.set_icon_tooltip_text(
                Gtk.EntryIconPosition.SECONDARY, None)

    def on_activate(self, widget, var_entry, val_entry, popover):
        if var_entry.get_icon_name(
                Gtk.EntryIconPosition.SECONDARY) != "gtk-apply":
            return
        if val_entry.get_icon_name(
                Gtk.EntryIconPosition.SECONDARY) != "gtk-apply":
            return

        var = var_entry.get_text().strip()
        val = val_entry.get_text().strip()

        self.emit("activated", var, val)

        popover.popdown()

    def on_closed(self, popover, var_entry, val_entry):
        var_entry.set_text("")
        val_entry.set_text("")


class CommandEditorCommandEntry(Gtk.MenuButton):
    __gsignals__ = {
        'activated': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self):
        super().__init__()

        self.set_label(_("Command"))
        self.set_mode(True)

        popover = Gtk.Popover(relative_to=self)

        grid = Gtk.Grid()
        grid.set_row_spacing(6)
        grid.set_column_spacing(6)
        grid.set_border_width(6)

        label = Gtk.Label.new("")
        label.set_xalign(0.00)
        label.set_markup("<b>%s</b>" % _("Command"))
        entry = Gtk.Entry()
        entry.connect('activate', self.on_command_entry_activate, popover)

        grid.attach(label, 0, 0, 1, 1)
        grid.attach(entry, 1, 0, 1, 1)

        bbox = Gtk.ButtonBox.new(Gtk.Orientation.HORIZONTAL)
        bbox.set_layout(Gtk.ButtonBoxStyle.START)
        bbox.set_spacing(6)
        bbox.set_homogeneous(True)
        grid.attach(bbox, 0, 1, 2, 1)

        file_chooser = Gtk.FileChooserButton.new(
            _("Select an executable"), Gtk.FileChooserAction.OPEN)
        file_chooser.connect('file-set', self.on_file_chooser_set, entry)
        filter = Gtk.FileFilter.new()
        filter.add_custom(Gtk.FileFilterFlags.FILENAME, self.exec_filter)
        file_chooser.set_filter(filter)
        bbox.pack_start(file_chooser, False, False, 0)

        app_chooser = AppChooserMenu()
        app_chooser.connect('selected', self.on_app_chooser_selected, entry)
        bbox.pack_start(app_chooser, False, False, 0)

        grid.show_all()

        popover.add(grid)

        popover.connect('closed', self.on_closed, entry)

        self.set_popover(popover)
        self.show_all()

    def exec_filter(self, filter_info):
        path = filter_info.filename
        return os.access(path, os.X_OK)

    def on_command_entry_activate(self, widget, popover):
        value = widget.get_text().strip()
        self.emit('activated', value)
        popover.popdown()

    def on_file_chooser_set(self, widget, entry):
        filename = widget.get_filename()
        entry.set_text(filename)
        entry.grab_focus()
        entry.set_position(len(filename))
        widget.set_filename("")

    def on_app_chooser_selected(self, widget, value, entry):
        entry.set_text(value)
        entry.grab_focus()
        entry.set_position(len(value))

    def on_closed(self, widget, entry):
        entry.set_text("")


class AppChooserMenu(Gtk.ComboBox):
    __gsignals__ = {
        'selected': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self):
        super().__init__()

        store = Gtk.ListStore.new([str, str, Gio.Icon])  # type: ignore
        store.append([_("Select an application"), None, None])

        app_list = []
        infos = Gio.AppInfo.get_all()
        for info in infos:
            executable = info.get_commandline()
            icon = info.get_icon()
            name = info.get_name()
            app_list.append([name, executable, icon])

        app_list = sorted(app_list, key=lambda x: x[0].lower())

        for app in app_list:
            store.append(app)

        self.set_model(store)

        renderer_pixbuf = Gtk.CellRendererPixbuf()
        self.pack_start(renderer_pixbuf, True)
        self.add_attribute(renderer_pixbuf, "gicon", 2)

        renderer_text = Gtk.CellRendererText()
        renderer_text.set_property('width-chars', 20)
        renderer_text.set_property('ellipsize', Pango.EllipsizeMode.END)
        self.pack_start(renderer_text, True)
        self.add_attribute(renderer_text, "text", 0)

        self.set_active(0)

        self.connect('changed', self.on_changed)

    def on_changed(self, widget):
        executable = self.get_executable()
        if executable is not None:
            self.emit('selected', executable)
        self.set_active(0)

    def get_executable(self):
        treemodel = self.get_model()
        treeiter = self.get_active_iter()
        if treeiter is None:
            return None
        row = treemodel[treeiter][:]  # type: ignore
        return row[1]


class FieldCodeMenu(Gtk.MenuButton):
    __gsignals__ = {
        'row-activated': (GObject.SignalFlags.RUN_FIRST, None,
                          (str,))
    }

    def __init__(self, label):
        super().__init__()

        self.set_label(label)
        self.set_mode(True)

        popover = Gtk.Popover(relative_to=self)

        self.listbox = Gtk.ListBox.new()
        popover.add(self.listbox)

        self.set_popover(popover)

        self.show_all()

        self.listbox.connect("row-activated", self.on_row_activated, popover)

    def add_option(self, value, text, size_group):
        option = FieldCodeOption(value, text, size_group)
        self.listbox.add(option)
        self.listbox.show_all()

    def on_row_activated(self, listbox, row, popover):
        self.emit("row-activated", row.get_value())
        popover.popdown()


class FieldCodeOption(Gtk.ListBoxRow):
    def __init__(self, value, text, size_group):
        super().__init__()

        box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        box.set_border_width(3)

        label = Gtk.Label.new("")
        label.set_markup("<i>%s</i>" % value)
        box.pack_start(label, False, False, 0)

        size_group.add_widget(label)

        label = Gtk.Label.new(text)
        label.set_markup(text)
        box.pack_start(label, False, False, 0)

        self.add(box)

        self.show_all()

        self.value = value

    def get_value(self):
        return self.value


class CommandEditorHint():
    def __init__(self, grid, row):
        self.image = Gtk.Image.new_from_icon_name(
            "gtk-apply", Gtk.IconSize.BUTTON)
        self.label = Gtk.Label.new("")
        self.label.set_xalign(0.00)
        self.label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
        grid.attach(self.image, 0, row, 1, 1)
        grid.attach(self.label, 1, row, 1, 1)

    def set_success(self, message):
        self.image.set_from_icon_name('gtk-apply', Gtk.IconSize.BUTTON)
        self.label.set_text(message)

    def set_error(self, message):
        self.image.set_from_icon_name('gtk-cancel', Gtk.IconSize.BUTTON)
        self.label.set_text(message)

    def set_state(self, state):
        if state[0]:
            self.set_success(state[1])
        else:
            self.set_error(state[1])


class CommandEntry(Gtk.Entry):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self):
        Gtk.Entry.__init__(self)

        self.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY,
            'document-edit-symbolic')

        self.connect("icon-release", self.on_icon_release)

        self.connect('changed', self._on_changed)

    def on_icon_release(self, entry, icon_pos, event):
        if icon_pos != Gtk.EntryIconPosition.SECONDARY:
            return

        dialog = CommandEditorDialog(
            self.get_toplevel(), self.get_text(), True)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self.set_text(dialog.get_text())

        dialog.destroy()

    def unescape_value(self, value):
        value = str(value)
        args = []
        for arg in shlex.split(value, posix=False):
            if arg.startswith("\""):
                arg = arg.replace("%%", "%")
            args.append(arg)
        return " ".join(args)

    def escape_value(self, value):
        value = str(value)
        args = []
        for arg in shlex.split(value, posix=False):
            if arg.startswith("\""):
                arg = arg.replace("%%", "%")  # Make it consistent for the pros
                arg = arg.replace("%", "%%")
            args.append(arg)
        return " ".join(args)

    def set_value(self, value):
        if value is None:
            value = ""
        self.set_text(self.unescape_value(value))

    def get_value(self):
        return self.escape_value(self.get_text())

    def _on_changed(self, widget):
        self.emit('value-changed', 'Exec', self.get_value())



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from .util import getDefaultMenuPrefix
from .Section import Section
from .AdvancedPage import AdvancedPage
from .ActionEditor import ActionEditor
from .CategoryEditor import CategoryEditor
from .TextEntryButton import TextEntryButton
from .SwitchEntry import SwitchEntry
from .PathEntry import PathEntry
from .FilenameLabel import FilenameLabel
from .IconEntry import IconEntry
from .CommandEditor import CommandEntry
from .MenulibreStackSwitcher import StackSwitcherBox
from .FieldLabel import FieldLabel
from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # type: ignore


class ApplicationEditor(Gtk.Box):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, use_headerbar):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        self._directory_hide_widgets = []
        self._internal_values = {
            'Type': '',
            'Version': ''
        }

        self.set_border_width(6)
        self.set_size_request(400, -1)

        scrolled = Gtk.ScrolledWindow.new(hadjustment=None, vadjustment=None)
        scrolled.set_shadow_type(Gtk.ShadowType.NONE)
        self.pack_start(scrolled, True, True, 0)

        vbox = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        scrolled.add(vbox)

        # Icon, Name, and Comment Entry
        hbox = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        vbox.add(hbox)

        self._icon_entry = IconEntry(use_headerbar=use_headerbar)
        self._icon_entry.connect("value-changed", self._on_changed)
        hbox.pack_start(self._icon_entry, False, False, 0)

        namebox = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        hbox.pack_start(namebox, True, True, 0)

        self._name_entry = TextEntryButton(
            property_name='Name',
            bold_font=True,
            required=True,
            placeholder_text=_('Add name'))
        self._name_entry.connect("value-changed", self._on_changed)
        namebox.pack_start(self._name_entry, True, True, 0)

        self._comment_entry = TextEntryButton(
            property_name='Comment', placeholder_text=_('Add comment'))
        self._comment_entry.connect("value-changed", self._on_changed)
        namebox.pack_start(self._comment_entry, True, True, 0)

        # Application Details Frame
        self._app_details = Section(_("Application Details"))
        vbox.add(self._app_details)

        self._directory_hide_widgets.append(self._app_details)

        grid = Gtk.Grid.new()
        grid.set_row_spacing(3)
        grid.set_column_spacing(12)
        self._app_details.add(grid)

        label = FieldLabel(
            _("Command"),
            "Exec",
            _(
                "Program to execute with arguments. This key is required if DBusActivatable is not set to \"True\" or if you need compatibility with implementations that do not understand D-Bus activation."),
            _("Supported Arguments"),
            "https://github.com/bluesabre/menulibre/wiki/Recognized-Desktop-Entry-Keys#exec")
        grid.attach(label, 0, 0, 1, 1)

        self._exec_entry = CommandEntry()
        self._exec_entry.connect("value-changed", self._on_changed)
        self._exec_entry.set_hexpand(True)
        grid.attach(self._exec_entry, 1, 0, 1, 1)

        label = FieldLabel(
            _("Working Directory"),
            "Path",
            _("The working directory."))
        grid.attach(label, 0, 1, 1, 1)

        self._path_entry = PathEntry(use_headerbar=use_headerbar)
        self._path_entry.connect("value-changed", self._on_changed)
        self._path_entry.set_hexpand(True)
        grid.attach(self._path_entry, 1, 1, 1, 1)

        # Options Frame
        self._options = Section(_("Options"))
        vbox.add(self._options)

        grid = Gtk.Grid.new()
        grid.set_row_spacing(3)
        grid.set_column_spacing(12)
        self._options.add(grid)

        label = FieldLabel(
            _("Run in terminal"),
            "Terminal",
            _("If set to \"True\", the program will be ran in a terminal window."))
        label.set_hexpand(True)
        grid.attach(label, 0, 0, 1, 1)

        self._terminal_entry = SwitchEntry('Terminal')
        self._terminal_entry.connect("value-changed", self._on_changed)
        grid.attach(self._terminal_entry, 1, 0, 1, 1)

        self._directory_hide_widgets.append(label)
        self._directory_hide_widgets.append(self._terminal_entry)

        label = FieldLabel(
            _("Use startup notification"),
            "StartupNotify",
            _("If set to \"True\", a startup notification is sent. Usually means that a busy cursor is shown while the application launches."))
        label.set_hexpand(True)
        grid.attach(label, 0, 1, 1, 1)

        self._startup_notify_entry = SwitchEntry('StartupNotify')
        self._startup_notify_entry.connect("value-changed", self._on_changed)
        grid.attach(self._startup_notify_entry, 1, 1, 1, 1)

        self._directory_hide_widgets.append(label)
        self._directory_hide_widgets.append(self._startup_notify_entry)

        label = FieldLabel(
            _("Hide from menus"),
            "NoDisplay",
            _("If set to \"True\", this entry will not be shown in menus, but will be available for MIME type associations etc."))
        label.set_hexpand(True)
        grid.attach(label, 0, 2, 1, 1)

        self._no_display_entry = SwitchEntry('NoDisplay')
        self._no_display_entry.connect("value-changed", self._on_changed)
        grid.attach(self._no_display_entry, 1, 2, 1, 1)

        # Settings Switcher
        self._additional_settings = StackSwitcherBox()
        vbox.pack_start(self._additional_settings, True, True, 0)

        self._directory_hide_widgets.append(self._additional_settings)

        # Categories Treeview and Inline Toolbar
        self._category_editor = CategoryEditor()
        self._category_editor.set_prefix(getDefaultMenuPrefix())
        self._category_editor.connect("value-changed", self._on_changed)
        self._additional_settings.add_child(self._category_editor,
                                            # Translators: "Categories"
                                            # launcher section
                                            'categories', _('Categories'))

        # Actions Treeview and Inline Toolbar
        self._action_editor = ActionEditor()
        self._action_editor.connect("value-changed", self._on_changed)
        self._additional_settings.add_child(self._action_editor,
                                            # Translators: "Actions" launcher
                                            # section
                                            'actions', _('Actions'))

        # Advanced Settings
        self._advanced_page = AdvancedPage(use_headerbar=use_headerbar)
        self._advanced_page.connect("value-changed", self._on_changed)
        self._additional_settings.add_child(self._advanced_page,
                                            # Translators: "Advanced" launcher
                                            # section
                                            'advanced', _('Advanced'))

        # Filename
        self._filename_label = FilenameLabel()
        self._filename_label.connect("value-changed", self._on_changed)
        vbox.pack_end(self._filename_label, False, False, 0)

    def set_value(self, key, value):
        """Set the DesktopSpec key, value pair in the editor."""
        if self._advanced_page.has_value(key):
            self._advanced_page.set_value(key, value)
        elif key == 'Filename':
            self._filename_label.set_value(value)
        elif key == 'Icon':
            self._icon_entry.set_value(value)
        elif key == 'Name':
            self._name_entry.set_value(value)
        elif key == 'Comment':
            self._comment_entry.set_value(value)
        elif key == 'Categories':
            self._category_editor.set_value(value)
        elif key == 'Actions':
            self._action_editor.set_value(value)
        elif key == 'Exec':
            self._exec_entry.set_value(value)
        elif key == 'Path':
            self._path_entry.set_value(value)
        elif key == 'Terminal':
            self._terminal_entry.set_value(value)
        elif key == 'StartupNotify':
            self._startup_notify_entry.set_value(value)
        elif key == 'NoDisplay':
            self._no_display_entry.set_value(value)
        elif key in self._internal_values.keys():
            self._internal_values[key] = value
        else:
            print("SET", key, value)

        if key == 'Type':
            for widget in self._directory_hide_widgets:
                if value == 'Directory':
                    widget.hide()
                else:
                    widget.show()

    def get_value(self, key):
        """Return the value stored for the specified key."""
        if self._advanced_page.has_value(key):
            return self._advanced_page.get_value(key)
        elif key == 'Name':
            return self._name_entry.get_value()
        elif key == 'Comment':
            return self._comment_entry.get_value()
        elif key == 'Icon':
            return self._icon_entry.get_value()
        elif key == 'Categories':
            return self._category_editor.get_value()
        elif key == 'Actions':
            return self._action_editor.get_value()
        elif key == 'Filename':
            return self._filename_label.get_value()
        elif key == 'Exec':
            return self._exec_entry.get_value()
        elif key == 'Path':
            return self._path_entry.get_value()
        elif key == 'Terminal':
            return self._terminal_entry.get_value()
        elif key == 'StartupNotify':
            return self._startup_notify_entry.get_value()
        elif key == 'NoDisplay':
            return self._no_display_entry.get_value()
        elif key in self._internal_values.keys():
            return self._internal_values[key]
        else:
            print("GET", key)
        return None

    def remove_incomplete_actions(self):
        self._action_editor.remove_incomplete_actions()

    def get_actions(self):
        return self._action_editor.get_actions()

    def clear_categories(self):
        self._category_editor.set_value("")

    def insert_required_categories(self, parent_directory):
        self._category_editor.insert_required_categories(parent_directory)

    def cancel(self):
        self._name_entry.cancel()
        self._comment_entry.cancel()

    def commit(self):
        self._name_entry.commit()
        self._comment_entry.commit()

    def take_focus(self):
        self._icon_entry.grab_focus()

    def _on_changed(self, widget, key, value):
        self.emit('value-changed', key, value)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Pango, GObject  # type: ignore


class TextEntryButton(Gtk.Stack):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(
            self,
            property_name,
            bold_font=False,
            required=False,
            placeholder_text=""):
        super().__init__()
        self._property_name = property_name
        self._value = ""
        self._required = required
        self._placeholder_text = placeholder_text
        self._bold_font = bold_font

        self._button = Gtk.Button.new()
        self._button.connect('clicked', self._on_clicked)
        self._button.set_relief(Gtk.ReliefStyle.NONE)
        self._button.connect('focus-in-event', self._on_button_focus_in)
        self._button.connect('focus-out-event', self._on_button_focus_out)
        self.add(self._button)

        self._label = Gtk.Label.new("")
        self._label.set_xalign(0.0)
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._button.add(self._label)

        self._entry = Gtk.Entry.new()
        self._entry.set_placeholder_text(placeholder_text)
        self._entry.connect('key-press-event', self._on_entry_key_press)
        self._entry.connect('activate', self._on_entry_activate)
        self._entry.connect('changed', self._on_entry_changed)
        self._entry.connect('icon-press', self._on_entry_icon_press)
        self.add(self._entry)

        self.set_homogeneous(True)

        self._on_entry_changed(self._entry)

    def set_value(self, value):
        if value is None:
            value = ""
        value = value.strip()
        self._value = value
        self._entry.set_text(value)
        attributes = Pango.AttrList.new()
        if len(value) > 0:
            self._label.set_text(value)
            attributes.insert(Pango.attr_style_new(Pango.Style.NORMAL))
            if self._bold_font:
                attributes.insert(Pango.attr_weight_new(Pango.Weight.BOLD))
                attributes.insert(Pango.attr_size_new(12000))
        else:
            self._label.set_text(self._placeholder_text)
            attributes.insert(Pango.attr_style_new(Pango.Style.ITALIC))
            attributes.insert(Pango.attr_weight_new(Pango.Weight.NORMAL))
        self._label.set_attributes(attributes)
        self.set_visible_child(self._button)
        self.emit('value-changed', self._property_name, self._value)

    def get_value(self):
        return self._entry.get_text()

    def _on_button_focus_in(self, button, event):
        button.set_relief(Gtk.ReliefStyle.NORMAL)

    def _on_button_focus_out(self, button, event):
        button.set_relief(Gtk.ReliefStyle.NONE)

    def _on_clicked(self, button):
        self.set_visible_child(self._entry)
        self._entry.grab_focus()

    def commit(self):
        if self._entry.get_icon_name(
                Gtk.EntryIconPosition.SECONDARY) == "gtk-apply":
            self.set_value(self._entry.get_text().strip())
        else:
            self.cancel()

    def cancel(self):
        self.set_visible_child(self._button)
        self._entry.set_text(self._value)

    def _on_entry_key_press(self, entry, event):
        keyval_name = Gdk.keyval_name(event.get_keyval()[1])
        if keyval_name is None:
            return
        if keyval_name.lower() == 'escape':
            self.cancel()
            self._button.grab_focus()

    def _on_entry_activate(self, entry):
        if entry.get_icon_name(Gtk.EntryIconPosition.SECONDARY) == "gtk-apply":
            self.commit()
            self._button.grab_focus()
        else:
            self.cancel()

    def _on_entry_changed(self, entry):
        text = entry.get_text().strip()
        if self._required and len(text) == 0:
            entry.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY,
                                          "gtk-cancel")
        else:
            entry.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY,
                                          "gtk-apply")

    def _on_entry_icon_press(self, entry, icon_pos, event):
        self._on_entry_activate(entry)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # type: ignore


COL_SHOW = 0
COL_NAME = 1
COL_DISPLAYED = 2
COL_COMMAND = 3


class ActionEditor(Gtk.Box):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=0)

        scrolled = Gtk.ScrolledWindow.new(hadjustment=None, vadjustment=None)
        scrolled.set_shadow_type(Gtk.ShadowType.IN)
        self.pack_start(scrolled, True, True, 0)

        self._liststore = Gtk.ListStore.new([bool, str, str, str])  # type: ignore

        treeview = Gtk.TreeView.new_with_model(self._liststore)
        treeview.set_enable_search(False)
        treeview.set_show_expanders(False)
        scrolled.add(treeview)

        renderer = Gtk.CellRendererToggle()
        renderer.set_property("activatable", True)
        renderer.connect("toggled", self._on_toggle_changed)

        # Translators: "Show" tree column header
        col = Gtk.TreeViewColumn(_("Show"), renderer, active=COL_SHOW)  # type: ignore
        treeview.append_column(col)

        renderer = Gtk.CellRendererText()
        renderer.set_property("editable", True)
        renderer.connect('edited', self._on_edited, COL_DISPLAYED)

        # Translators: "Name" tree column header
        col = Gtk.TreeViewColumn(_("Name"), renderer, text=COL_DISPLAYED)  # type: ignore
        treeview.append_column(col)

        renderer = Gtk.CellRendererText()
        renderer.set_property("editable", True)
        renderer.connect('edited', self._on_edited, COL_COMMAND)

        # Translators: "Command" tree column header
        col = Gtk.TreeViewColumn(_("Command"), renderer, text=COL_COMMAND)  # type: ignore
        treeview.append_column(col)

        toolbar = Gtk.Toolbar.new()
        context = toolbar.get_style_context()
        context.add_class("inline-toolbar")
        self.add(toolbar)

        add_button = self._make_button(_("Add"), "list-add-symbolic")
        add_button.connect("clicked", self._on_add_clicked)
        toolbar.add(add_button)

        remove_button = self._make_button(
            _("Remove"), "list-remove-symbolic", False)
        remove_button.connect("clicked", self._on_remove_clicked, treeview)
        toolbar.add(remove_button)

        clear_button = self._make_button(
            _("Clear"), "list-remove-all-symbolic", False)
        clear_button.connect("clicked", self._on_clear_clicked)
        toolbar.add(clear_button)

        up_button = self._make_button(_("Move Up"), "go-up-symbolic", False)
        up_button.connect("clicked", self._on_move_clicked, treeview, -1)
        toolbar.add(up_button)

        down_button = self._make_button(
            _("Move Down"), "go-down-symbolic", False)
        down_button.connect("clicked", self._on_move_clicked, treeview, 1)
        toolbar.add(down_button)

        treeview.connect(
            "cursor-changed",
            self._on_cursor_changed,
            remove_button,
            up_button,
            down_button)

        self._row_change_inhibit = False
        self._row_change_singleton = False
        self._liststore.connect("row-changed", self._on_row_changed)
        self._liststore.connect(
            "row-inserted",
            self._on_row_inserted,
            treeview,
            remove_button,
            clear_button,
            up_button,
            down_button)
        self._liststore.connect(
            "row-deleted",
            self._on_row_deleted,
            treeview,
            remove_button,
            clear_button,
            up_button,
            down_button)

        self.show_all()

    def remove_incomplete_actions(self):
        rows = []
        for row in self._liststore:
            if len(row[COL_NAME]) > 0 and len(row[COL_COMMAND]) > 0:
                rows.append(row[:])

        self._liststore.clear()

        for row in rows:
            self._liststore.append(row)

    def _has_prev_path(self, model, path):
        if path.copy().prev():
            return True
        return False

    def _has_next_path(self, model, path):
        try:
            current_path_str = path.to_string()
            next_path_str = str(int(current_path_str) + 1)
            next_path = Gtk.TreePath.new_from_string(next_path_str)
            model.get_iter(next_path)
        except (TypeError, ValueError):
            return False
        return True

    def _toggle_move_buttons(
            self,
            treeview,
            remove_button,
            up_button,
            down_button):
        try:
            model, [path] = treeview.get_selection().get_selected_rows()
        except ValueError:
            model = path = None

        can_up = False
        can_down = False
        can_remove = False

        if path:
            can_up = self._has_prev_path(model, path)
            can_down = self._has_next_path(model, path)
            can_remove = True

        up_button.set_sensitive(can_up)
        down_button.set_sensitive(can_down)
        remove_button.set_sensitive(can_remove)

    def _on_cursor_changed(
            self,
            treeview,
            remove_button,
            up_button,
            down_button):
        self._toggle_move_buttons(
            treeview, remove_button, up_button, down_button)

    def _on_toggle_changed(self, cell, path):
        treeiter = self._liststore.get_iter(path)
        self._liststore.set_value(treeiter, COL_SHOW, not cell.get_active())
        self.emit('value-changed', 'Actions', self.get_value())

    def _on_edited(self, widget, row, new_text, col):
        """Edited callback function to enable modifications to a cell."""
        self._liststore[row][col] = new_text
        self.emit('value-changed', 'Actions', self.get_value())

    def _on_move_clicked(self, widget, treeview, relative_position=0):
        sel = treeview.get_selection().get_selected()
        if sel:
            model, selected_iter = sel

            # Move the row up if relative_position < 0
            if relative_position < 0:
                sibling = model.iter_previous(selected_iter)
                if sibling is not None:
                    model.move_before(selected_iter, sibling)
            else:
                sibling = model.iter_next(selected_iter)
                if sibling is not None:
                    model.move_after(selected_iter, sibling)
        self.emit('value-changed', 'Actions', self.get_value())

    def _make_button(self, label, icon_name, sensitive=True):
        image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
        image.set_pixel_size(16)
        button = Gtk.ToolButton.new(image, label)
        button.set_sensitive(sensitive)
        return button

    def _set_actions(self, rows):
        self._row_change_inhibit = True

        self._clear()

        for row in rows:
            self._liststore.append(row)

        self._row_change_inhibit = False

    def set_value(self, value):
        if value is None:
            value = []
        else:
            value = json.loads(value)
        self._set_actions(value)

    def get_actions(self):
        model = self._liststore
        results = []
        for row in model:
            results.append(row[:])
        return results

    def get_value(self):
        return json.dumps(self.get_actions())

    def _on_row_changed(self, model, path, treeiter):
        if self._row_change_singleton:
            self._row_change_singleton = False
            return
        self._row_change_singleton = True
        self.emit('value-changed', 'Actions', self.get_value())

    def _on_row_inserted(
            self,
            model,
            path,
            treeiter,
            treeview,
            remove_button,
            clear_button,
            up_button,
            down_button):
        if not self._row_change_inhibit:
            treeview.set_cursor(path, treeview.get_column(0), True)
        clear_button.set_sensitive(True)
        self._toggle_move_buttons(
            treeview, remove_button, up_button, down_button)

    def _on_row_deleted(
            self,
            model,
            path,
            treeview,
            remove_button,
            clear_button,
            up_button,
            down_button):
        sensitive = len(self._liststore) > 0
        clear_button.set_sensitive(sensitive)
        self._toggle_move_buttons(
            treeview, remove_button, up_button, down_button)

    def _clear(self):
        self._liststore.clear()
        self._removals = []

    def _on_add_clicked(self, widget):
        existing = list()
        for row in self._liststore:
            existing.append(row[COL_NAME])
        name = 'NewShortcut'
        n = 1
        while name in existing:
            name = 'NewShortcut%i' % n
            n += 1
        # Translators: Placeholder text for a newly created action
        displayed = _("New Shortcut")
        self._liststore.append([True, name, displayed, ''])
        self.emit('value-changed', 'Actions', self.get_value())

    def _on_remove_clicked(self, widget, treeview):
        model, treeiter = treeview.get_selection().get_selected()
        if model is not None and treeiter is not None:
            self._removals.append(model[treeiter][0])
            model.remove(treeiter)
            self.emit('value-changed', 'Actions', self.get_value())

    def _on_clear_clicked(self, widget):
        self._clear()
        self.emit('value-changed', 'Actions', self.get_value())



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # type: ignore


class Headerbar(Gtk.HeaderBar):

    def __init__(self):
        super().__init__()

        self.set_title("MenuLibre")
        self.set_custom_title(Gtk.Label.new())
        self.set_show_close_button(True)

    def add_menu_button(self, icon_name, label, menu):
        image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        image.set_pixel_size(16)
        image.set_property('use-fallback', True)

        button = Gtk.MenuButton.new()
        button.set_menu_model(menu)
        button.set_use_popover(True)
        button.set_tooltip_text(label)
        button.add(image)

        self.add(button)

        return button

    def add_button(self, icon_name, label):
        item = Gtk.Button.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        item.set_tooltip_text(label)

        image = item.get_image()
        if image is not None:
            image.set_property('use-fallback', True)

        self.add(item)

        return item

    def add_buttons(self, button_specs):
        box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)

        context = box.get_style_context()
        context.add_class("linked")

        items = []

        for icon_name, label in button_specs:
            item = Gtk.Button.new_from_icon_name(
                icon_name, Gtk.IconSize.BUTTON)
            item.set_tooltip_text(label)
            image = item.get_image()
            if image is not None:
                image.set_property('use-fallback', True)
            box.add(item)
            items.append(item)

        self.add(box)

        return items

    def add_search(self, widget):
        # Need to insert on right side
        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.set_valign(Gtk.Align.CENTER)

        box.add(widget)
        self.pack_end(box)

        return box



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from .TextEntry import TextEntry
from .StartupWmClassEntry import StartupWmClassEntry
from .SwitchEntry import SwitchEntry
from .FieldLabel import FieldLabel
from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # type: ignore


class AdvancedPage(Gtk.ScrolledWindow):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, use_headerbar):
        super().__init__(hadjustment=None, vadjustment=None)  # type: ignore

        self._row_index = 0
        self._widgets = {}

        self.set_shadow_type(Gtk.ShadowType.IN)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self._grid = Gtk.Grid()
        self._grid.set_border_width(12)
        self._grid.set_row_spacing(3)
        self._grid.set_row_homogeneous(True)
        self._grid.set_column_spacing(6)
        self._grid.set_valign(Gtk.Align.START)
        self.add(self._grid)

        self._add_row(
            'GenericName',
            _("Generic Name"),
            _('Generic name of the application, for example "Web Browser".'),
            TextEntry('GenericName'))

        self._add_row(
            'TryExec',
            _("Try Exec"),
            _('Path to an executable file to determine if the program is installed. If the file is not present or is not executable, this entry may not be shown in a menu.'),
            TextEntry('TryExec'))

        self._add_row(
            'OnlyShowIn', _("Only Show In"), _(
                'A list of environments that should display this entry. Other environments will not display this entry. You can only use this key if "NotShowIn" is not set.\n'
                'Possible values include: Budgie, Cinnamon, EDE, GNOME, KDE, LXDE, LXQt, MATE, Pantheon, Razor, ROX, TDE, Unity, XFCE, Old'), TextEntry('OnlyShowIn'))

        self._add_row(
            'NotShowIn',
            _("Not Shown In"),
            _(
                'A list of environments that should not display this entry. You can only use this key if "OnlyShowIn" is not set.\n'
                'Possible values include: Budgie, Cinnamon, EDE, GNOME, KDE, LXDE, LXQt, MATE, Pantheon, Razor, ROX, TDE, Unity, XFCE, Old'),
            TextEntry('NotShowIn'))

        self._add_row('MimeType',
                      _("Mimetypes"),
                      _('The MIME type(s) supported by this application.'),
                      TextEntry('MimeType'))

        self._add_row(
            'Keywords',
            _("Keywords"),
            _('A list of keywords to describe this entry. You can use these to help searching entries. These are not meant for display, and should not be redundant with the values of Name or GenericName.'),
            TextEntry('Keywords'))

        self._add_row(
            'StartupWMClass',
            _("Startup WM Class"),
            _('If specified, the application will be requested to use the string as a WM class or a WM name hint at least in one window.'),
            StartupWmClassEntry(
                'StartupWMClass',
                use_headerbar=use_headerbar))

        self._add_row(
            'Implements',
            _("Implements"),
            _('A list of interfaces that this application implements.'),
            TextEntry('Implements'))

        self._add_row(
            'Hidden',
            _("Hidden"),
            _('If set to "True", the result for the user is equivalent to the .desktop file not existing at all.'),
            SwitchEntry('Hidden'))

        self._add_row(
            'DBusActivatable',
            _("DBUS Activatable"),
            _(
                'Set this key to "True" if D-Bus activation is supported for this application and you want to use it.'),
            SwitchEntry('DBusActivatable'),
            _("More Information"),
            "https://github.com/bluesabre/menulibre/wiki/Recognized-Desktop-Entry-Keys#dbusactivatable")

        self._add_row(
            'PrefersNonDefaultGPU',
            _("Prefers Non-Default GPU"),
            _(
                'Set this key to "True" if this application prefers to be run on a more powerful GPU if available.'),
            SwitchEntry('PrefersNonDefaultGPU'),
            _("More Information"),
            "https://github.com/bluesabre/menulibre/wiki/Recognized-Desktop-Entry-Keys#prefersnondefaultgpu")

        self._add_row(
            'X-GNOME-UsesNotifications',
            _("Uses Notifications"),
            _(
                'Set this key to "True" if this application uses notifications.'),
            SwitchEntry('X-GNOME-UsesNotifications'),
            _("More Information"),
            "https://github.com/bluesabre/menulibre/wiki/Recognized-Desktop-Entry-Keys#x-gnome-usesnotifications")

        self.show_all()

    def _add_row(self, id, text, tooltip, widget,
                 help_text=None, help_url=None):
        label = FieldLabel(
            label=text, key_name=id, description=tooltip, help_text=help_text,
            help_url=help_url
        )
        self._grid.attach(label, 0, self._row_index, 1, 1)

        widget.connect('value-changed', self._on_widget_value_changed)
        self._grid.attach(widget, 1, self._row_index, 1, 1)
        self._widgets[id] = widget

        self._row_index += 1

    def _on_widget_value_changed(self, widget, property_name, value):
        self.emit('value-changed', property_name, value)

    def has_value(self, property_name):
        return property_name in list(self._widgets.keys())

    def set_value(self, property_name, value):
        if self.has_value(property_name):
            self._widgets[property_name].set_value(value)
            return True
        return False

    def get_value(self, property_name):
        if self.has_value(property_name):
            return self._widgets[property_name].get_value()
        return None



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk  # type: ignore


class StackSwitcherBox(Gtk.Box):
    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=6)

        self._stack = Gtk.Stack()
        self._stack.set_transition_type(Gtk.StackTransitionType.NONE)
        self._stack.set_transition_duration(500)

        self._switcher = Gtk.StackSwitcher()
        self._switcher.set_stack(self._stack)
        self._switcher.set_property("valign", Gtk.Align.CENTER)
        self._switcher.set_property("halign", Gtk.Align.CENTER)

        self.pack_start(self._switcher, False, False, 0)
        self.pack_start(self._stack, True, True, 0)

    def get_stack(self):
        return self._stack

    def get_switcher(self):
        return self._switcher

    def add_child(self, child, name, title):
        self._stack.add_titled(child, name, title)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _
import os

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Pango, GObject, GdkPixbuf  # type: ignore


class IconEntry(Gtk.MenuButton):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, use_headerbar):
        super().__init__()

        self._value = ""

        menu = Gtk.Menu.new()
        self.set_popup(menu)

        item = Gtk.MenuItem.new_with_label(_("Browse Icons…"))
        item.connect("activate", self._on_browse_icons_clicked, use_headerbar)
        menu.append(item)

        item = Gtk.MenuItem.new_with_label(_("Browse Files…"))
        item.connect("activate", self._on_browse_files_clicked, use_headerbar)
        menu.append(item)

        menu.show_all()

        overlay = Gtk.Overlay.new()
        overlay.set_size_request(48, 48)
        self.add(overlay)

        self._image = Gtk.Image.new_from_icon_name(
            "folder", Gtk.IconSize.DIALOG)
        self._image.set_pixel_size(48)
        overlay.add(self._image)

        self._overlay_image = Gtk.Image.new_from_icon_name(
            "dialog-warning", Gtk.IconSize.MENU)
        self._overlay_image.set_pixel_size(16)
        self._overlay_image.set_halign(Gtk.Align.END)
        self._overlay_image.set_valign(Gtk.Align.END)
        self._overlay_image.set_no_show_all(True)
        overlay.add_overlay(self._overlay_image)
        overlay.set_overlay_pass_through(self._overlay_image, True)

    def _on_browse_icons_clicked(self, widget, use_headerbar):
        dialog = IconSelectionDialog(self.get_toplevel(), "", use_headerbar)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self._set_image(dialog.get_icon())

        dialog.destroy()

    def _on_browse_files_clicked(self, widget, use_headerbar):
        dialog = IconFileSelectionDialog(
            self.get_toplevel(), "", use_headerbar)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            self._set_image(dialog.get_filename())

        dialog.destroy()

    def set_value(self, value):
        self._set_image(value)

    def _set_image(self, icon_name):
        # Load the Icon Theme.
        icon_theme = Gtk.IconTheme.get_default()

        if icon_name is not None:
            # If the Icon Theme has the icon, set the image to that icon.
            if icon_theme.has_icon(icon_name):
                self._image.set_from_icon_name(icon_name, 48)
                self._on_successful_image(icon_name, icon_name)
                return

            # If the Icon Theme has a symbolic version of the icon, set the
            # image to that icon.
            if icon_theme.has_icon(icon_name + "-symbolic"):
                icon_name = icon_name + "-symbolic"
                self._image.set_from_icon_name(icon_name, 48)
                self._on_successful_image(icon_name, icon_name)
                return

            # If the icon name is actually a file, render it to the Image.
            elif os.path.isfile(icon_name):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(icon_name)
                if pixbuf is not None:
                    scaled = pixbuf.scale_simple(48, 48,
                                                GdkPixbuf.InterpType.HYPER)
                    self._image.set_from_pixbuf(scaled)
                    self._on_successful_image(icon_name, icon_name)
                    return

            if icon_name.startswith("applications-"):
                replacement_icon_name = "folder"
            elif icon_name == "gnome-settings":
                replacement_icon_name = "folder"
            else:
                replacement_icon_name = "application-x-executable"

            self._image.set_from_icon_name(replacement_icon_name, 48)
            self._on_error_image(
                icon_name,
                _("<i>Missing icon:</i> %s") %
                icon_name)
            return

        if icon_theme.has_icon("applications-other"):
            icon_name = "applications-other"
        else:
            icon_name = "application-x-executable"

        self._image.set_from_icon_name(icon_name, 48)
        self._on_successful_image(icon_name, icon_name)

    def get_value(self):
        return self._value

    def _on_successful_image(self, icon_name, tooltip_text):
        self.set_tooltip_text(tooltip_text)
        self._image.set_opacity(1.0)
        self._overlay_image.hide()
        self._value = icon_name
        self.emit('value-changed', 'Icon', icon_name)

    def _on_error_image(self, icon_name, tooltip_markup):
        self.set_tooltip_markup(tooltip_markup)
        self._image.set_opacity(0.7)
        self._overlay_image.show()
        self._value = icon_name
        self.emit('value-changed', 'Icon', icon_name)


class IconFileSelectionDialog(Gtk.FileChooserDialog):

    def __init__(self, parent, initial_file="", use_header_bar=False):
        super().__init__(title=_("Select an image"), transient_for=parent,
                         action=Gtk.FileChooserAction.OPEN,
                         use_header_bar=use_header_bar)
        self.add_buttons(
            _('Cancel'), Gtk.ResponseType.CANCEL,
            _('Apply'), Gtk.ResponseType.OK
        )

        file_filter = Gtk.FileFilter()
        # Translators: "Images" file chooser dialog filter
        file_filter.set_name(_("Images"))
        file_filter.add_mime_type("image/*")
        self.add_filter(file_filter)

        if len(initial_file) > 0:
            self.set_filename(initial_file)


class IconSelectionDialog(Gtk.Dialog):

    def __init__(self, parent, initial_icon="", use_header_bar=False):
        super().__init__(title=_("Select an icon"), transient_for=parent,
                         use_header_bar=use_header_bar, flags=0)  # type: ignore
        self.add_buttons(
            _('Cancel'), Gtk.ResponseType.CANCEL,
            _('Apply'), Gtk.ResponseType.OK
        )

        self.set_default_size(640, 480)
        self.set_default_response(Gtk.ResponseType.OK)

        box = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        box.set_margin_top(9)
        box.set_margin_bottom(12)
        box.set_margin_start(12)
        box.set_margin_end(12)

        search = Gtk.SearchEntry.new()
        box.pack_start(search, False, False, 0)

        scrolled = Gtk.ScrolledWindow.new(None, None)
        scrolled.set_shadow_type(Gtk.ShadowType.IN)
        box.pack_start(scrolled, True, True, 0)

        self.treeview = IconSelectionTreeView(search)
        self.treeview.connect("row-activated", self._on_treeview_row_activated)
        scrolled.add(self.treeview)

        search.connect(
            "search-changed",
            self._on_search_changed,
            self.treeview)
        self.connect("key-press-event", self._on_key_press_event, search)

        self.get_content_area().pack_start(box, True, True, 0)
        self.show_all()

        if len(initial_icon) > 0:
            self.treeview.set_icon(initial_icon)

        self.treeview.grab_focus()

    def _on_search_changed(self, widget, treeview):
        treeview.refilter()
        treeview.set_cursor(Gtk.TreePath.new_first())

    def _on_treeview_row_activated(self, tree_view, path, column):
        self.response(Gtk.ResponseType.OK)

    def _on_key_press_event(self, widget, event, search):
        if not bool(event.get_state() & Gdk.ModifierType.CONTROL_MASK):
            return False
        keyval_name = Gdk.keyval_name(event.get_keyval()[1])
        if keyval_name is None:
            return False
        if keyval_name.lower() != 'f':
            return False
        search.grab_focus()

    def get_icon(self):
        return self.treeview.get_icon()


class IconSelectionTreeView(Gtk.TreeView):
    def __init__(self, search_entry):
        super().__init__()

        icons = Gtk.ListStore.new([str, str])  # type: ignore

        icon_theme = Gtk.IconTheme.get_default()
        icons_list = icon_theme.list_icons(None)
        icons_list = sorted(icons_list, key=lambda x: x[0].lower())

        for icon_name in icons_list:
            icons.append([icon_name.lower(), icon_name])

        self.filter = icons.filter_new()
        self.filter.set_visible_func(self._icon_sel_match_func, search_entry)

        self.set_model(Gtk.TreeModelSort(model=self.filter))
        self.set_headers_visible(False)
        self.set_enable_search(False)

        icon_col = Gtk.TreeViewColumn.new()
        icon_col.set_sort_column_id(0)
        renderer_pixbuf = Gtk.CellRendererPixbuf()
        renderer_pixbuf.set_property('width', 32)
        renderer_pixbuf.set_property('height', 32)
        icon_col.pack_start(renderer_pixbuf, False)
        icon_col.add_attribute(renderer_pixbuf, "icon-name", 1)
        self.append_column(icon_col)

        text_col = Gtk.TreeViewColumn.new()
        text_col.set_sort_column_id(0)
        renderer_text = Gtk.CellRendererText()
        renderer_text.set_property('ellipsize', Pango.EllipsizeMode.END)
        text_col.pack_start(renderer_text, True)
        text_col.add_attribute(renderer_text, "text", 1)
        self.append_column(text_col)
        text_col.emit("clicked")

    def refilter(self):
        self.filter.refilter()

    def _icon_sel_match_func(self, model, treeiter, entry):
        query = str(entry.get_text().lower())

        if query == "":
            return True

        return query in model[treeiter][0].lower()

    def _get_path_to_icon(self, icon_name):
        model = self.get_model()
        if model is None:
            return None
        idx = 0
        treeiter = model.get_iter_first()
        while treeiter is not None:
            if model[treeiter][0] == icon_name:
                return Gtk.TreePath.new_from_string(str(idx))
            treeiter = model.iter_next(treeiter)
            idx = idx + 1
        return None

    def set_icon(self, icon_name):
        path = self._get_path_to_icon(icon_name.lower())
        if path is not None:
            self.set_cursor(path)
        else:
            self.set_cursor(Gtk.TreePath.new_first())

    def get_icon(self):
        model, treeiter = self.get_selection().get_selected()
        if treeiter is None:
            return None
        return model[treeiter][0]



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2016 OmegaPhil <OmegaPhil@startmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from locale import gettext as _

from gi.repository import Gio, GObject, Gtk, Pango, GLib  # type: ignore

from . import MenuEditor, MenulibreXdg, XmlMenuElementTree, util
from .util import MenuItemTypes, check_keypress, getBasename, getRelativeName, escapeText

import logging
logger = logging.getLogger('menulibre')


class Treeview(Gtk.Box):

    __gsignals__ = {
        'cursor-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
                           (GObject.TYPE_BOOLEAN,)),
        'add-directory-enabled': (GObject.SIGNAL_RUN_LAST,
                                  GObject.TYPE_BOOLEAN,
                                  (GObject.TYPE_BOOLEAN,)),
        'requires-menu-reload': (GObject.SIGNAL_RUN_LAST,
                                 GObject.TYPE_BOOLEAN,
                                 (GObject.TYPE_BOOLEAN,)),
    }

    loaded = False

    def __init__(self, parent):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.parent = parent
        self._last_selected_path = -1
        self._search_terms = None
        self._lock_menus = False

        self.set_size_request(220, -1)

        scrolled = Gtk.ScrolledWindow.new(hadjustment=None, vadjustment=None)
        scrolled.set_shadow_type(Gtk.ShadowType.IN)
        scrolled.set_name("MenulibreSidebarScroll")
        self.pack_start(scrolled, True, True, 0)

        treestore = MenuEditor.get_treestore()
        if treestore is not None:
            self.loaded = True
            self._treeview = Gtk.TreeView.new_with_model(treestore)
        else:
            self._treeview = Gtk.TreeView.new()

        self._treeview.set_show_expanders(True)
        self._treeview.set_enable_search(False)
        self._treeview.set_headers_visible(False)
        scrolled.add(self._treeview)

        # Translators: "Search Results" treeview column header
        col = Gtk.TreeViewColumn(_("Search Results"))  # type: ignore

        # Icon renderer
        col_cell_img = Gtk.CellRendererPixbuf()
        col_cell_img.set_property("stock-size", Gtk.IconSize.LARGE_TOOLBAR)
        col.pack_start(col_cell_img, False)

        # Text renderer
        col_cell_text = Gtk.CellRendererText()
        col_cell_text.set_property("ellipsize", Pango.EllipsizeMode.END)
        col.pack_start(col_cell_text, True)

        # Set the markup property on the Text cell.
        col.add_attribute(col_cell_text, "markup", MenuEditor.COL_DISPLAY_NAME)

        # Add the cell data func for the text column to render labels.
        col.set_cell_data_func(col_cell_text, self._text_display_func, None)

        # Set the Tooltip column.
        self._treeview.set_tooltip_column(1)

        # Add the cell data func for the pixbuf column to render icons.
        col.set_cell_data_func(col_cell_img, self._icon_name_func, None)

        # Append the column, set the model.
        self._treeview.append_column(col)

        # Configure the treeview events.
        self._treeview.connect("cursor-changed",
                               self._on_treeview_cursor_changed, None)
        self._treeview.connect("key-press-event",
                               self._on_treeview_key_press_event, None)
        self._treeview.connect("row-expanded",
                               self._on_treeview_row_expansion, True)
        self._treeview.connect("row-collapsed",
                               self._on_treeview_row_expansion, False)

        self.menu_timeout_id = 0

        self._toolbar = Gtk.Toolbar.new()
        self._toolbar.set_icon_size(Gtk.IconSize.MENU)
        self._toolbar.set_name("MenulibreSidebarToolbar")
        self.add(self._toolbar)

        context = self._toolbar.get_style_context()
        context.add_class("inline-toolbar")

        img = Gtk.Image.new_from_icon_name('go-up-symbolic', Gtk.IconSize.MENU)
        self._move_up_button = Gtk.ToolButton.new(img, _("Move Up"))
        self._move_up_button.set_tooltip_text(_("Move Up"))
        self._move_up_button.connect(
            'clicked', self._move_iter, (self._treeview, -1))
        self._toolbar.add(self._move_up_button)

        img = Gtk.Image.new_from_icon_name(
            'go-down-symbolic', Gtk.IconSize.MENU)
        self._move_down_button = Gtk.ToolButton.new(img, _("Move Down"))
        self._move_down_button.set_tooltip_text(_("Move Down"))
        self._move_down_button.connect(
            'clicked', self._move_iter, (self._treeview, 1))
        self._toolbar.add(self._move_down_button)

        img = Gtk.Image.new_from_icon_name(
            'view-sort-ascending-symbolic', Gtk.IconSize.MENU)
        self._sort_button = Gtk.ToolButton.new(img, _("Sort Alphabetically"))
        self._sort_button.set_tooltip_text(_("Sort Alphabetically"))
        self._sort_button.connect('clicked', self._sort_iter)
        self._toolbar.add(self._sort_button)

        # Show the treeview, grab focus.
        self.show_all()
        self._treeview.grab_focus()

    def set_sortable(self, sortable):
        self._sort_button.set_sensitive(sortable)

    def set_move_up_enabled(self, enabled):
        self._move_up_button.set_sensitive(enabled)

    def set_move_down_enabled(self, enabled):
        self._move_down_button.set_sensitive(enabled)

# TreeView Modifiers
    def append(self, row_data):
        """Add a new launcher entry below the current selected one."""
        model, treeiter = self._get_selected_iter()
        model, parent = self.get_parent()

        new_iter = model.insert_after(parent, treeiter)
        self._populate_and_select_iter(model, new_iter, row_data)

        return new_iter

    def prepend(self, row_data):
        """Add a new launcher entry above the current selected one."""
        model, treeiter = self._get_selected_iter()
        parent = self.get_parent()

        new_iter = model.insert_before(parent, treeiter)
        self._populate_and_select_iter(model, new_iter, row_data)

        return new_iter

    def add_child(self, row_data, treeiter=None, model=None, do_select=True):
        """Add a new child launcher to the current selected one, or the
        specified iter if adding elsewhere in the tree, with optional
        selection."""
        if treeiter is None or model is None:
            model, treeiter = self._get_selected_iter()

        if treeiter is None:
            return None

        new_iter = model.prepend(treeiter)

        if do_select:
            self._treeview.expand_row(model[treeiter].path, False)

        self._populate_and_select_iter(model, new_iter, row_data, do_select)

        return new_iter

    def remove_selected(self, ui_only=False):  # noqa
        """Remove the selected launcher. If ui_only is True, the associated
        desktop file is not deleted but the launcher is removed from the
        interface. This is useful for removing categories."""

        self._last_selected_path = -1
        model, treeiter = self._get_selected_iter()

        filename = None
        original = None
        item_type = None
        del_files = []

        if not ui_only:

            if treeiter is not None:
                filename = model[treeiter][MenuEditor.COL_FILENAME]
                item_type = model[treeiter][MenuEditor.COL_TYPE]

            if filename is not None:
                basename = getRelativeName(filename)
                original = util.getSystemLauncherPath(basename)
            else:
                original = None

            # Get files for deletion - only one item can be selected at a time,
            # but this may be a directory
            del_dirs, del_apps = self._get_delete_filenames(model, treeiter)
            del_files = del_dirs + del_apps

            # Uninstall the launcher
            self.xdg_menu_uninstall(model, treeiter, filename)

            # Delete each of the files - this will fail silently for non-user
            # desktop files/directories, these are hidden below
            for filename in del_files:
                try:
                    os.remove(filename)
                except:  # noqa
                    pass

        self.xdg_menu_update()
        self._cleanup_applications_merged()

        if not ui_only:

            # How is it possible that a launcher is to be removed, but its
            # associated desktop file has not been returned to be deleted?
            if filename is None or filename not in del_files:
                # Update the required categories.
                model, parent_data = self.get_parent_row_data()
                if parent_data is not None:
                    categories = util.getRequiredCategories(
                        parent_data[MenuEditor.COL_FILENAME])
                else:
                    categories = util.getRequiredCategories(None)
                self.parent.update_launcher_categories(categories, [])

            if original is not None:
                # Original found (this is a system-installed desktop file/
                # directory rather than a user-made one), hide the desktop file
                # and all associated instances in the menu
                entry = MenulibreXdg.MenulibreDesktopEntry(original)
                name = entry['Name']
                comment = entry['Comment']
                categories = entry['Categories']
                executable = entry['Exec']
                icon_name = entry['Icon']
                hidden = entry['Hidden'] or entry['NoDisplay']
                self.update_selected(
                    name,
                    comment,
                    executable,
                    categories,
                    item_type,
                    icon_name,
                    original,
                    not hidden)
                model, row_data = self.get_selected_row_data()
                self.update_launcher_instances(filename, row_data)
                treeiter = None

        if treeiter is not None:
            path = model.get_path(treeiter)
            if model is not None and treeiter is not None:
                if not isinstance(model, Gtk.TreeModelFilter):
                    model.remove(treeiter)
            if path:
                self._treeview.set_cursor(path)

        self.update_menus()

    def remove_iter(self, model, treeiter):
        """Remove launcher pointed to by iter from the model only - use this
        when you need to remove a launcher from a non-selected directory, e.g.
        when the associated category has been removed"""

        # This feels a bit redundant for a function, but it keeps the
        # functionality close to remove_selected
        model.remove(treeiter)

# Get
    def get_parent(self, model=None, treeiter=None):
        """Get the parent iterator for the current treeiter"""
        parent = None
        if model is None:
            model, treeiter = self._get_selected_iter()
        if treeiter is not None:
            path = model.get_path(treeiter)
            if path.up():
                if path.get_depth() > 0:
                    try:
                        parent = model.get_iter(path)
                    except:  # noqa
                        parent = None
        return model, parent

    def is_first(self, model=None, treeiter=None):
        if model is None:
            model, treeiter = self._get_selected_iter()
        if treeiter:
            path = model.get_path(treeiter)
            if path.prev():
                return False
        return True

    def _next(self, model, treeiter, path):
        # path.next() seems to be broken, so we will do it ourselves.
        try:
            string = path.to_string()
            parts = string.split(":")
            parts[-1] = str(int(parts[-1]) + 1)
            string = ":".join(parts)
            path = Gtk.TreePath.new_from_string(string)
            model.get_iter(path)
        except (TypeError, ValueError):
            return None
        return path

    def is_last(self, model=None, treeiter=None):
        if model is None:
            model, treeiter = self._get_selected_iter()
        if treeiter:
            path = model.get_path(treeiter)

            if self._next(model, treeiter, path) is not None:
                return False
        return True

    def get_parent_filename(self):
        """Get the filename of the parent iter."""
        model, parent = self.get_parent()
        if parent is None:
            return None
        return model[parent][MenuEditor.COL_FILENAME]

    def get_parent_row_data(self):
        """Get the row data of the parent iter."""
        model, parent = self.get_parent()
        if parent is None:
            return model, None
        parent_data: list = model[parent][:]  # type: ignore
        return model, parent_data

    def get_selected_filename(self):
        """Return the filename of the current selected treeiter."""
        model, row_data = self.get_selected_row_data()
        if row_data is not None:
            return row_data[MenuEditor.COL_FILENAME]
        return None

    def get_selected_row_data(self):
        """Get the row data of the current selected item."""
        model, treeiter = self._get_selected_iter()
        if model is None or treeiter is None:
            return model, None
        row_data: list = model[treeiter][:]  # type: ignore
        return model, row_data

# Set
    def set_can_select_function(self, can_select_func):
        """Set the external function used for can-select."""
        selection = self._treeview.get_selection()
        selection.set_select_function(self._on_treeview_selection,
                                      can_select_func)

# Update
    def update_launcher_instances(self, filename, row_data):
        """Update all same launchers with the new information."""
        model, treeiter = self._get_selected_iter()
        for instance in self._get_launcher_instances(filename, model):
            for i in range(len(row_data)):
                model[instance][i] = row_data[i]

    def update_selected(self, name, comment, executable, categories, item_type,
                        icon_name, filename, show=True):
        """Update the application treeview selected row data."""
        model, treeiter = self._get_selected_iter()
        if treeiter is None:
            return

        model[treeiter][MenuEditor.COL_NAME] = name
        model[treeiter][MenuEditor.COL_DISPLAY_NAME] = escapeText(name)
        model[treeiter][MenuEditor.COL_COMMENT] = comment
        model[treeiter][MenuEditor.COL_EXEC] = executable
        model[treeiter][MenuEditor.COL_CATEGORIES] = categories
        model[treeiter][MenuEditor.COL_TYPE] = item_type
        if os.path.isfile(icon_name):
            gfile = Gio.File.parse_name(icon_name)
            icon = Gio.FileIcon.new(gfile)
        else:
            icon = Gio.ThemedIcon.new(icon_name)
        model[treeiter][MenuEditor.COL_G_ICON] = icon
        model[treeiter][MenuEditor.COL_ICON_NAME] = icon_name
        model[treeiter][MenuEditor.COL_FILENAME] = filename
        model[treeiter][MenuEditor.COL_SHOW] = show

        # Refresh the displayed launcher
        self._last_selected_path = -1
        self._on_treeview_cursor_changed(self._treeview, None)

# Events
    def _on_treeview_cursor_changed(self, widget, selection):
        """Update the editor frame when the selected row is changed."""
        # Check if the selection is valid.
        sel = widget.get_selection()
        if sel:
            treestore, treeiter = sel.get_selected()
            if not treestore:
                return
            if not treeiter:
                return

            # Do nothing if we didn't change path
            path = str(treestore.get_path(treeiter))
            if path == self._last_selected_path:
                return
            self._last_selected_path = path

            # Notify the application that the cursor selection has changed.
            self.emit("cursor-changed", True)

            # Update the Add Directory menu item
            self._update_add_directory()

    def _on_treeview_key_press_event(self, widget, event, user_data=None):
        """Handle treeview keypress events."""
        # Right expands the selected row.
        if check_keypress(event, ['right']):
            self._set_treeview_selected_expanded(widget, True)
            return True
        # Left collapses the selected row.
        elif check_keypress(event, ['left']):
            self._set_treeview_selected_expanded(widget, False)
            return True
        # Spacebar toggles the expansion of the selected row.
        elif check_keypress(event, ['space']):
            self._toggle_treeview_selected_expanded(widget)
            return True
        return False

    def _on_treeview_row_expansion(self, treeview, treeiter, column, expanded):
        if self._toolbar.get_sensitive():
            model = treeview.get_model()
            row = model[treeiter]
            row[MenuEditor.COL_EXPAND] = expanded

    def _on_treeview_selection(self, sel, store, path, is_selected,
                               can_select_func):
        """Save changes on cursor change."""
        if is_selected:
            return can_select_func()
        return True

# Helper functions
    def _set_treeview_selected_expanded(self, treeview, expanded=True):
        """Set the expansion (True or False) of the selected row."""
        sel = treeview.get_selection()
        model, treeiter = sel.get_selected()
        row = model[treeiter]
        if expanded:
            treeview.expand_row(row.path, False)
        else:
            treeview.collapse_row(row.path)

    def _toggle_treeview_selected_expanded(self, treeview):
        """Toggle the expansion of the selected row."""
        expanded = self._get_treeview_selected_expanded(treeview)
        self._set_treeview_selected_expanded(treeview, not expanded)

    def _text_display_func(self, col, renderer, treestore, treeiter,
                           user_data):
        """CellRenderer function to set the gicon for each row."""
        show = treestore[treeiter][MenuEditor.COL_SHOW]
        if show:
            renderer.set_property("style", Pango.Style.NORMAL)
        else:
            renderer.set_property("style", Pango.Style.ITALIC)
        separator = treestore[treeiter][MenuEditor.COL_TYPE] == MenuItemTypes.SEPARATOR  # type: ignore
        renderer.set_property("sensitive", not separator)
        renderer.set_property("style-set", True)

    def _icon_name_func(self, col, renderer, treestore, treeiter, user_data):
        """CellRenderer function to set the gicon for each row."""
        renderer.set_property("gicon",
                              treestore[treeiter][MenuEditor.COL_G_ICON])
        separator = treestore[treeiter][MenuEditor.COL_TYPE] == MenuItemTypes.SEPARATOR  # type: ignore
        renderer.set_property("sensitive", not separator)

    def _get_selected_iter(self):
        """Return the current treeview model and selected iter."""
        model: Gtk.TreeStore
        model, treeiter = self._treeview.get_selection().get_selected()  # type: ignore
        return model, treeiter

    def _populate_and_select_iter(self, model, treeiter, row_data,
                                  do_select=True):
        """Fill the specified treeiter with data and optionally select it."""
        for i in range(len(row_data)):
            model[treeiter][i] = row_data[i]

        # Select the new iter if requested
        if do_select:
            path = model.get_path(treeiter)
            self._treeview.set_cursor(path)

    def _get_deletable_launcher(self, filename):
        """Return True if the launcher is available for deletion."""
        if not os.path.exists(filename):
            return False
        return True

    def _get_delete_filenames(self, model, treeiter):  # noqa
        """Return a list of files to be deleted after uninstall."""
        directories = []
        applications = []

        filename = model[treeiter][MenuEditor.COL_FILENAME]
        block_run = False

        if filename is not None:
            basename = getRelativeName(filename)
            original = util.getSystemLauncherPath(basename)
            item_type = model[treeiter][MenuEditor.COL_TYPE]
            if original is None and item_type == MenuItemTypes.DIRECTORY:  # type: ignore
                pass
            else:
                block_run = True

        if model.iter_has_child(treeiter) and not block_run:
            for i in range(model.iter_n_children(treeiter)):
                child_iter = model.iter_nth_child(treeiter, i)
                filename = model[child_iter][MenuEditor.COL_FILENAME]
                if filename is not None:
                    if filename.endswith('.directory'):
                        d, a = self._get_delete_filenames(model, child_iter)
                        directories = directories + d
                        applications = applications + a
                        directories.append(filename)
                    else:
                        if self._get_deletable_launcher(filename):
                            applications.append(filename)

        filename = model[treeiter][MenuEditor.COL_FILENAME]
        if filename is not None:
            if filename.endswith('.directory'):
                directories.append(filename)
            else:
                if self._get_deletable_launcher(filename):
                    applications.append(filename)
        return directories, applications

    def _get_treeview_selected_expanded(self, treeview):
        """Return True if the selected row is currently expanded."""
        sel = treeview.get_selection()
        model, treeiter = sel.get_selected()
        row = model[treeiter]
        return treeview.row_expanded(row.path)

    def _get_launcher_instances(self, filename, model=None, parent=None):
        """Return a list of all treeiters referencing this filename."""
        if filename is None:
            return []
        if model is None:
            model, treeiter = self._get_selected_iter()
        treeiters = []
        for n_child in range(model.iter_n_children(parent)):
            treeiter = model.iter_nth_child(parent, n_child)
            if treeiter is None:
                continue
            iter_filename = model[treeiter][MenuEditor.COL_FILENAME]
            if iter_filename == filename:
                treeiters.append(treeiter)
            if model.iter_has_child(treeiter):
                treeiters += self._get_launcher_instances(filename, model,
                                                          treeiter)
        return treeiters

    def _get_n_launcher_instances(self, filename):
        return len(self._get_launcher_instances(filename))

    def _is_menu_locked(self):
        """Return True if menu editing is currently locked."""
        return self._lock_menus

    def _update_add_directory(self):
        """Prevent adding subdirectories to system menus."""
        add_enabled = True
        prefix = util.getDefaultMenuPrefix()

        treestore, treeiter = self._get_selected_iter()
        model, parent_iter = self.get_parent()
        while parent_iter is not None:
            filename = treestore[parent_iter][MenuEditor.COL_FILENAME]
            if getBasename(filename).startswith(prefix):
                add_enabled = False
            model, parent_iter = self.get_parent(treestore, parent_iter)

        self.emit("add-directory-enabled", add_enabled)

# Search
    def search(self, terms):
        """Search the treeview for the specified terms."""
        self._search_terms = str(terms.lower())
        model: Gtk.TreeModelFilter = self._treeview.get_model()  # type: ignore
        model.refilter()

    def set_searchable(self, searchable, expand=False):
        """Set the TreeView searchable."""
        model = self._treeview.get_model()
        if model is None:
            return

        if searchable:
            self._lock_menus = True
            # Show the "Search Results" header and disable the inline toolbar.
            self._treeview.set_headers_visible(True)
            self._toolbar.set_sensitive(False)

            # If specified, expand the treeview.
            if expand:
                self._treeview.expand_all()

            # If the model is not a filter, make it so.
            if not isinstance(model, Gtk.TreeModelFilter):
                model = model.filter_new()
                self._treeview.set_model(model)
                model.set_visible_func(self._treeview_match_func)

        else:
            self._lock_menus = False
            # Hide the headers and enable the inline toolbar.
            self._treeview.set_headers_visible(False)
            self._toolbar.set_sensitive(True)

            if isinstance(model, Gtk.TreeModelFilter):
                # Get the model and iter.
                f_model, f_iter = self._get_selected_iter()

                # Restore the original model.
                model = model.get_model()
                self._treeview.set_model(model)

                # Restore expanded items (lp 1307000)
                self._treeview.collapse_all()
                for n_child in range(model.iter_n_children(None)):
                    treeiter = model.iter_nth_child(None, n_child)
                    if treeiter is None:
                        continue
                    row = model[treeiter]
                    if row[MenuEditor.COL_FILENAME]:
                        self._treeview.expand_row(row.path, False)

                # Try to get the row that was selected previously.
                if (f_model is not None) and (f_iter is not None):
                    row_data = f_model[f_iter][:]  # type: ignore
                    selected_iter = self._get_iter_by_data(row_data, model,
                                                           parent=None)

                # If that fails, just select the first iter.
                else:
                    selected_iter = model.get_iter_first()

                if selected_iter is not None:
                    # Set the cursor.
                    path = model.get_path(selected_iter)
                    self._treeview.set_cursor(path)

    def _treeview_match(self, model, treeiter, query):
        """Match subfunction for filtering search results."""
        name = model[treeiter][MenuEditor.COL_NAME]
        comment = model[treeiter][MenuEditor.COL_COMMENT]
        executable = model[treeiter][MenuEditor.COL_EXEC]
        item_type = model[treeiter][MenuEditor.COL_TYPE]
        desktop = model[treeiter][MenuEditor.COL_FILENAME]

        # Hide separators in the search results.
        if item_type == MenuItemTypes.SEPARATOR:  # type: ignore
            return False

        # Convert None to blank.
        if not name:
            name = ""
        if not comment:
            comment = ""
        if not executable:
            executable = ""

        # Expand all the rows.
        self._treeview.expand_all()

        # Match against the name.
        if query in name.lower():
            return True

        # Match against the comment.
        if query in comment.lower():
            return True

        if query in executable.lower():
            return True

        # Match against the desktop file.
        desktop = desktop.replace("menulibre-", "")
        desktop = desktop.replace("alacarte-", "")
        if query in desktop.lower():
            return True

        # Show the directory if any child items match.
        if item_type == MenuItemTypes.DIRECTORY:  # type: ignore
            return self._treeview_match_directory(query, model, treeiter)

        # No matches, return False.
        return False

    def _treeview_match_directory(self, query, model, treeiter):
        """Match subfunction for matching directory children."""
        for child_i in range(model.iter_n_children(treeiter)):
            child = model.iter_nth_child(treeiter, child_i)
            if self._treeview_match(model, child, query):
                return True

        return False

    def _treeview_match_func(self, model, treeiter, data=None):
        """Match function for filtering search results."""
        # Make the query case-insensitive.
        if self._search_terms == "":
            return True

        return self._treeview_match(model, treeiter, self._search_terms)

# XDG Menu Commands
    def xdg_menu_install(self, filename, parent=None):
        """Install the specified filename in the menu structure."""
        model, treeiter = self._get_selected_iter()
        if filename is None:
            return
        if treeiter is None:
            return
        if filename.endswith('.desktop'):
            menu_install = True
            parents = []
            if parent is None:
                parent = model.iter_parent(treeiter)
            while parent is not None:
                parent_filename = model[parent][MenuEditor.COL_FILENAME]

                # Do not do this method if this is a known system directory.
                parents.append(parent_filename)
                parent = model.iter_parent(parent)
            parents.reverse()
            if menu_install:
                installed = MenulibreXdg.desktop_menu_install(parents, [
                                                              filename])
                if not installed:
                    self.emit("requires-menu-reload", True)

    def xdg_menu_uninstall(self, model, treeiter, filename):
        """Uninstall the specified filename from the menu structure."""
        if filename is None:
            return
        if filename.endswith('.desktop'):
            menu_install = True
            menu_prefix = util.getDefaultMenuPrefix()
            parents = []
            parent = model.iter_parent(treeiter)
            while parent is not None:
                parent_filename = model[parent][MenuEditor.COL_FILENAME]
                # Do not do this method if this is a known system directory.
                if getBasename(parent_filename).startswith(menu_prefix):
                    menu_install = False
                parents.append(parent_filename)
                parent = model.iter_parent(parent)
            parents.reverse()
            if menu_install:
                MenulibreXdg.desktop_menu_uninstall(parents, [filename])

    def xdg_menu_update(self):
        """Force an update of the Xdg Menu."""
        MenulibreXdg.desktop_menu_update()

    def update_menus(self):
        """Update the menu files."""
        if self.menu_timeout_id > 0:
            GLib.source_remove(self.menu_timeout_id)
        self.menu_timeout_id = GLib.timeout_add_seconds(
            1, self.update_menu_timeout)

    def update_menu_timeout(self):
        # Do not save menu layout if in search mode (lp #1306999)
        if not self._is_menu_locked():
            XmlMenuElementTree.treeview_to_xml(self._treeview)
            self.update_menus_kde()
        self.menu_timeout_id = 0
        return False

    def update_menus_kde(self):
        try:
            subprocess.Popen(["kbuildsycoca5"])
        except FileNotFoundError:
            pass

    def _cleanup_applications_merged(self):
        """Cleanup items from ~/.config/menus/applications-merged"""
        # xdg-desktop-menu installs menu files in
        # ~/.config/menus/applications-merged, but does not remove them
        # correctly.
        merged_dir = os.path.join(GLib.get_user_config_dir(),
                                  "menus", "applications-merged")

        # Get the list of installed user directories to compare with.
        directories_dir = os.path.join(GLib.get_home_dir(),
                                       ".local", "share",
                                       "desktop-directories")
        if os.path.isdir(directories_dir):
            directories = os.listdir(directories_dir)
        else:
            directories = []

        # Check if applications-merged actually exists...
        if os.path.isdir(merged_dir):
            for menufile in os.listdir(merged_dir):
                menufile = os.path.join(merged_dir, menufile)
                remove_file = False

                # Only interested in .menu files
                if os.path.isfile(menufile) and menufile.endswith('.menu'):
                    logger.debug("Checking if %s is still valid..." %
                                 menufile)

                    # Read the menufile to see if it has a valid directory.
                    with open(menufile) as menufile_tmp:
                        for line in menufile_tmp.readlines():
                            if "<Directory>" in line:
                                menuname = line.split('<Directory>')[1]
                                menuname = menuname.split('</Directory>')[0]
                                menuname = menuname.strip()

                                # Remove if a listed directory is not installed
                                if menuname not in directories:
                                    remove_file = True
                    if remove_file:
                        logger.debug("Removing useless %s" % menufile)
                        os.remove(menufile)

# TreeView iter tricks
    def _move_iter(self, widget, user_data):  # noqa
        """Move the currently selected row up or down. If the neighboring row
        is expanded, make the selected row a child of the neighbor row.

        Keyword arguments:
        widget -- the triggering GtkWidget
        user_data -- list-packed parameters:
            treeview -- the GtkTreeview being modified
            relative_position -- 1 or -1, determines moving up or down

        """
        # Unpack the user data
        treeview, relative_position = user_data

        # Get the current selected row
        sel = treeview.get_selection().get_selected()
        if sel:
            model, selected_iter = sel
            selected_type = model[selected_iter][MenuEditor.COL_TYPE]

            # Get current required categories
            model, parent = self.get_parent(model, selected_iter)
            if parent:
                categories = util.getRequiredCategories(
                    model[parent][MenuEditor.COL_FILENAME])
            else:
                categories = util.getRequiredCategories(None)

            # Move the row up if relative_position < 0
            if relative_position < 0:
                sibling_iter = model.iter_previous(selected_iter)
            else:
                sibling_iter = model.iter_next(selected_iter)

            if sibling_iter:
                sibling_path = model.get_path(sibling_iter)

                # Determine where the item is being inserted.
                move_down = False

                # What is the neighboring item?
                sibling_type = model[sibling_iter][MenuEditor.COL_TYPE]

                # Sibling Directory
                if sibling_type == MenuItemTypes.DIRECTORY:  # type: ignore
                    # Do not move directories into other directories.
                    if selected_type == MenuItemTypes.DIRECTORY:  # type: ignore
                        move_down = False

                    # Append or Prepend to expanded directories.
                    elif treeview.row_expanded(sibling_path):
                        move_down = True

                    # Append to childless directories (lp: #1318209)
                    elif not model.iter_has_child(sibling_iter):
                        move_down = True

                # Insert the selected item into the directory.
                if move_down:
                    selected_iter = self._move_iter_down_level(
                        treeview, selected_iter,
                        sibling_iter, relative_position)

                # Move the selected item before or after the sibling item.
                else:
                    if relative_position < 0:
                        model.move_before(selected_iter, sibling_iter)
                    else:
                        model.move_after(selected_iter, sibling_iter)

            # If there is no neighboring row, move up a level.
            else:
                selected_iter = self._move_iter_up_level(
                    treeview, selected_iter, relative_position)

            # Get new required categories
            model, parent = self.get_parent(model, selected_iter)
            if parent:
                new_categories = util.getRequiredCategories(
                    model[parent][MenuEditor.COL_FILENAME])
            else:
                new_categories = util.getRequiredCategories(None)

            # Replace required categories
            if categories != new_categories:
                editor_categories = self.parent.get_editor_categories()
                split_categories = editor_categories.split(';')
                for category in categories:
                    if category in split_categories:
                        split_categories.remove(category)
                for category in new_categories:
                    if category not in split_categories:
                        split_categories.append(category)
                split_categories.sort()
                editor_categories = ';'.join(split_categories)
                self.parent.set_editor_categories(editor_categories)
                self.parent.update_launcher_categories(categories,
                                                       new_categories)

        self.update_menus()

        self.scroll_to_selection()

        self.emit("cursor-changed", True)

    def _get_iter_by_data(self, row_data, model: Gtk.TreeModel, parent=None):
        """Search the TreeModel for a row matching row_data.

        Return the TreeIter found or None if none found."""
        for n_child in range(model.iter_n_children(parent)):
            treeiter = model.iter_nth_child(parent, n_child)
            if treeiter is None:
                continue
            if model[treeiter][:] == row_data:  # type: ignore
                return treeiter
            if model.iter_n_children(treeiter) != 0:
                value = self._get_iter_by_data(row_data, model, treeiter)
                if value is not None:
                    return value
        return None

    def _move_iter_up_level(self, treeview, treeiter, relative_position):
        """Move the specified iter up one level."""
        model = treeview.get_model()
        sibling = model.iter_parent(treeiter)
        if sibling is not None:
            parent = model.iter_parent(sibling)
            row_data = model[treeiter][:]
            if relative_position < 0:
                new_iter = model.insert_before(parent,
                                               sibling,
                                               row_data)
            else:
                new_iter = model.insert_after(parent,
                                              sibling,
                                              row_data)

            # Install/Uninstall items from directories.
            filename = row_data[MenuEditor.COL_FILENAME]
            self.xdg_menu_install(filename)
            self.xdg_menu_uninstall(model, treeiter, filename)

            model.remove(treeiter)
            path = model.get_path(new_iter)
            treeview.set_cursor(path)
            return new_iter

    def _move_iter_down_level(self, treeview, treeiter, parent_iter,
                              relative_position):
        """Move the specified iter down one level."""
        model = treeview.get_model()
        row_data = model[treeiter][:]
        if model.iter_has_child(parent_iter):
            if relative_position < 0:
                n_children = model.iter_n_children(parent_iter)
                sibling = model.iter_nth_child(parent_iter, n_children - 1)
                new_iter = model.insert_after(parent_iter, sibling, row_data)
            else:
                sibling = model.iter_nth_child(parent_iter, 0)
                new_iter = model.insert_before(parent_iter, sibling, row_data)
        else:
            new_iter = model.insert(parent_iter, 0, row_data)

        # Install/Uninstall items from directories.
        filename = row_data[MenuEditor.COL_FILENAME]
        self.xdg_menu_install(filename, parent_iter)
        self.xdg_menu_uninstall(model, treeiter, filename)

        model.remove(treeiter)
        treeview.expand_row(model[parent_iter].path, False)
        path = model.get_path(new_iter)
        treeview.set_cursor(path)
        return new_iter

    def _sort_iter(self, widget):
        """Alphabetical sort of items in the current directory."""

        # Get the current selected row
        model, sel_iter = self._get_selected_iter()
        if sel_iter is not None:

            # Move to the parent iter - if there is no parent, it must be the
            # top level, which is ignored
            item_names = []
            _, parent_iter = self.get_parent(model, sel_iter)
            if parent_iter:

                # Deteriming list of item names
                for i in range(model.iter_n_children(parent_iter)):
                    child_iter = model.iter_nth_child(parent_iter, i)
                    if child_iter is not None:
                        item_names.append(model[child_iter][MenuEditor.COL_NAME])

                # Applying unstable (?) case-insensitive alphabetical sort
                item_names = sorted(item_names, key=str.lower)

                for i in range(len(item_names)):
                    child_iter = model.iter_nth_child(parent_iter, i)
                    if child_iter is None:
                        continue

                    # Ignore if item is already sorted or at least has an
                    # identical title to that expected
                    if item_names[i] != model[child_iter][MenuEditor.COL_NAME]:
                        search_iter = None

                        # Locating desired item in the remaining unsorted items
                        for r in range(i, len(item_names)):
                            search_iter = model.iter_nth_child(parent_iter, r)
                            if search_iter is None:
                                continue
                            if item_names[i] == model[search_iter][MenuEditor.COL_NAME]:
                                break

                        if search_iter is not None:
                            # Moving the found item into place
                            model.move_before(search_iter, child_iter)

                # Committing changes
                self.update_menus()

            self.scroll_to_selection()

    def scroll_to_selection(self):
        model, sel_iter = self._get_selected_iter()
        if sel_iter is None:
            return
        self._treeview.scroll_to_cell(model.get_path(sel_iter), None,
                                      False, 0.0, 0.0)

    def set_search_entry(self, entry):
        self._treeview.set_search_entry(entry)

    def reset_cursor(self):
        self._treeview.set_cursor(Gtk.TreePath.new_from_string("1"))
        self._treeview.set_cursor(Gtk.TreePath.new_from_string("0"))



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#   Copyright (C) 2017-2018 OmegaPhil <OmegaPhil@startmail.com>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango, GObject  # type: ignore


try:
    from . import FileHandler
except ImportError:
    pass


class ParsingErrorsDialog(Gtk.Dialog):

    def __init__(self, parent, use_header_bar=False, demo_mode=False):
        super().__init__(title=_("Parsing Errors"), transient_for=parent,
                         use_header_bar=use_header_bar, flags=0)  # type: ignore
        self.add_buttons(
            _('Close'), Gtk.ResponseType.CLOSE,
        )

        self.set_default_size(900, 480)
        # self.set_default_size(1240, 480)

        box = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        box.set_margin_top(9)
        box.set_margin_bottom(12)
        box.set_margin_start(12)
        box.set_margin_end(12)

        message_area = Gtk.Box.new(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        box.pack_start(message_area, False, False, 0)

        image = Gtk.Image.new_from_icon_name(
            "dialog-warning", Gtk.IconSize.DIALOG)
        message_area.pack_start(image, False, False, 0)

        message = Gtk.Label.new(_(
            "The following desktop files have failed parsing by the underlying library, and will therefore not show up in MenuLibre.\n"
            "Please investigate these problems with the associated package maintainer."
        ))
        message.set_line_wrap(True)
        message.set_xalign(0.0)
        message_area.pack_start(message, True, True, 0)

        scrolled = Gtk.ScrolledWindow.new(None, None)
        scrolled.set_shadow_type(Gtk.ShadowType.IN)
        box.pack_start(scrolled, True, True, 0)

        self.listbox = Gtk.ListBox.new()
        scrolled.add(self.listbox)

        self.demo_mode = demo_mode
        if self.demo_mode:
            self.file_handler = None
        else:
            self.file_handler = FileHandler.FileHandler()

        self.get_content_area().pack_start(box, True, True, 0)
        self.show_all()

    def add_item(self, filename, error):
        row = ErrorRow(filename, error)
        row.connect("action", self.on_row_action)
        self.listbox.add(row)

    def on_row_action(self, row, action, filename):
        if self.demo_mode:
            print(action, filename)
            return
        if self.file_handler is None:
            return
        if action == "open-folder":
            self.file_handler.open_folder(filename)
        elif action == "open-editor":
            self.file_handler.open_editor(filename)
        elif action == "copy-location":
            self.file_handler.copy_to_clipboard(filename)


class ErrorRow(Gtk.ListBoxRow):
    __gsignals__ = {
        'action': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, filename, error):
        super().__init__()

        box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        box.set_border_width(3)
        box.set_margin_start(3)
        box.set_margin_end(3)

        label = Gtk.Label.new("")
        label.set_xalign(0.0)
        label.set_ellipsize(Pango.EllipsizeMode.START)
        label.set_markup("<b>%s</b>\n%s" % (filename, error))
        box.pack_start(label, True, True, 0)

        buttons = Gtk.ButtonBox.new(orientation=Gtk.Orientation.HORIZONTAL)
        buttons.set_layout(Gtk.ButtonBoxStyle.END)
        context = buttons.get_style_context()
        context.add_class("linked")
        box.pack_end(buttons, False, False, 0)

        button = Gtk.Button.new_from_icon_name(
            "folder-symbolic", Gtk.IconSize.BUTTON)
        button.set_tooltip_text(_("Open containing folder"))
        button.connect("clicked", self.emit_action, filename, "open-folder")
        buttons.add(button)

        button = Gtk.Button.new_from_icon_name(
            "text-editor-symbolic", Gtk.IconSize.BUTTON)
        button.set_tooltip_text(_("Open in text editor"))
        button.connect("clicked", self.emit_action, filename, "open-editor")
        buttons.add(button)

        button = Gtk.Button.new_from_icon_name(
            "edit-copy-symbolic", Gtk.IconSize.BUTTON)
        button.set_tooltip_text(_("Copy file location"))
        button.connect("clicked", self.emit_action, filename, "copy-location")
        buttons.add(button)

        self.add(box)

        self.show_all()

    def emit_action(self, button, filename, action):
        self.emit("action", action, filename)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject  # type: ignore


class PathEntry(Gtk.Entry):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self, use_headerbar):
        super().__init__()

        self._value = ""

        self.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, "folder-open-symbolic")

        self.connect("changed", self._on_entry_changed)
        self.connect("icon-release", self._on_icon_clicked, use_headerbar)

    def set_value(self, value):
        if value is None:
            value = ""
        self.set_text(value)

    def get_value(self):
        return self.get_text()

    def _on_entry_changed(self, widget):
        self.emit('value-changed', 'Path', self.get_value())

    def _on_icon_clicked(self, entry, icon_pos, event, use_headerbar):
        """Show the file selection dialog when Path Browse is clicked."""
        # Translators: File Chooser Dialog, window title.
        title = _("Select a working directory")
        action = Gtk.FileChooserAction.SELECT_FOLDER

        dialog = FileChooserDialog(
            self.get_toplevel(), title, action, use_headerbar)
        dialog.set_filename(self.get_value())
        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
            self.set_value(filename)
        dialog.destroy()
        entry.grab_focus()


class FileChooserDialog(Gtk.FileChooserDialog):
    def __init__(self, parent, title, action, use_headerbar):
        Gtk.FileChooserDialog.__init__(
            self,
            title=title,
            transient_for=parent,
            action=action,
            use_header_bar=use_headerbar)
        # Translators: File Chooser Dialog, cancel button.
        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        # Translators: File Chooser Dialog, confirmation button.
        self.add_button(_("OK"), Gtk.ResponseType.OK)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk  # type: ignore


# lint:disable
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
# lint:enable


def set_up_logging(opts):
    """Set up logging for menulibre"""
    # add a handler to prevent basicConfig
    root = logging.getLogger()
    null_handler = NullHandler()
    root.addHandler(null_handler)

    formatter = logging.Formatter("%(levelname)s:%(name)s: "
                                  "%(funcName)s() '%(message)s'")

    logger = logging.getLogger('menulibre')
    logger_sh = logging.StreamHandler()
    logger_sh.setFormatter(formatter)
    logger.addHandler(logger_sh)

    lib_logger = logging.getLogger('menulibre_lib')
    lib_logger_sh = logging.StreamHandler()
    lib_logger_sh.setFormatter(formatter)
    lib_logger.addHandler(lib_logger_sh)

    # Set the logging level to show debug messages.
    try:
        if opts.verbose:
            logger.setLevel(logging.DEBUG)
            logger.debug('logging enabled')
        if opts.verbose > 1:
            lib_logger.setLevel(logging.DEBUG)
    except TypeError:
        pass


def show_uri(parent, link):
    """Open a web browser to the specified link."""
    screen = parent.get_screen()
    Gtk.show_uri(screen, link, Gtk.get_current_event_time())


def alias(alternative_function_name):
    '''see http://www.drdobbs.com/web-development/184406073#l9'''
    def decorator(function):
        '''attach alternative_function_name(s) to function'''
        if not hasattr(function, 'aliases'):
            function.aliases = []
        function.aliases.append(alternative_function_name)
        return function
    return decorator



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

import optparse
import sys

from locale import gettext as _

from menulibre import MenulibreApplication

from menulibre_lib import set_up_logging, get_version


def parse_options():
    """Support for command line options"""
    parser = optparse.OptionParser(version="%%prog %s" % get_version())
    parser.add_option(
        "-v", "--verbose", action="count", dest="verbose",
        # Translators: Command line option to display debug messages on stdout
        help=_("Show debug messages"))
    parser.add_option(
        "-b", "--headerbar", action="count", dest="headerbar",
        # Translators: Command line option to switch layout
        help=_("Use headerbar layout (client side decorations)")
    )
    parser.add_option(
        "-t", "--toolbar", action="count", dest="toolbar",
        # Translators: Command line option to switch layout
        help=_("Use toolbar layout (server side decorations)")
    )
    (options, args) = parser.parse_args()

    set_up_logging(options)

    return options


def main():
    """Main application for Menulibre"""
    opts = parse_options()

    # Run the application.
    app = MenulibreApplication.Application()
    if opts.headerbar is not None:
        app.use_headerbar = True
    elif opts.toolbar is not None:
        app.use_toolbar = True

    exit_status = app.run(None)
    sys.exit(exit_status)



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#   MenuLibre - Advanced fd.o Compliant Menu Editor
#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
#
#   This program is free software: you can redistribute it and/or modify it
#   under the terms of the GNU General Public License version 3, as published
#   by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranties of
#   MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
#   PURPOSE.  See the GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program.  If not, see <http://www.gnu.org/licenses/>.

from locale import gettext as _
from .LabelWithHidingButton import LabelWithHidingButton

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, Pango, GObject  # type: ignore

class FilenameLabel(LabelWithHidingButton):
    __gsignals__ = {
        'value-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, str,)),
    }

    def __init__(self):
        super().__init__(
            label="",
            icon_name="edit-copy-symbolic",
            icon_size=Gtk.IconSize.BUTTON)
        self._value = None

        label = self.get_label()
        label.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
        label.set_xalign(0.0)
        label.set_yalign(1.0)

        attributes = Pango.AttrList.new()
        attributes.insert(Pango.attr_style_new(Pango.Style.ITALIC))
        attributes.insert(Pango.attr_weight_new(Pango.Weight.NORMAL))
        label.set_attributes(attributes)

        button = self.get_button()
        button.set_tooltip_text(_("Copy"))
        button.connect("clicked", self._button_clicked_cb)

    def set_value(self, value):
        if value is None or value == "":
            value = None
            text = ""
        else:
            text = value
        self._label.set_text(text)
        self._label.set_tooltip_text(text)
        self._value = value
        self.emit('value-changed', 'Filename', self._value)

    def get_value(self):
        return self._value

    def _button_clicked_cb(self, widget):
        if self._value is None:
            return
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(self._value, -1)
